# -*- coding: utf-8 -*-
import math
from abjad import Fraction
from abjad.tools import durationtools
from abjad.tools import mathtools
from abjad.tools import systemtools
from abjad.tools.scoretools.Container import Container
from abjad.tools.topleveltools import inspect_
from abjad.tools.topleveltools import iterate
from abjad.tools.topleveltools import mutate
from abjad.tools.topleveltools import override
from abjad.tools.topleveltools import select


class Tuplet(Container):
    r'''A tuplet.

    ..  container:: example

        **Example 1.** A tuplet:

            >>> tuplet = Tuplet(Multiplier(2, 3), "c'8 d'8 e'8")
            >>> show(tuplet) # doctest: +SKIP

        ..  doctest::

            >>> print(format(tuplet))
            \times 2/3 {
                c'8
                d'8
                e'8
            }

    ..  container:: example

        **Example 2.** A nested tuplet:

            >>> second_tuplet = Tuplet((4, 7), "g'4. ( a'16 )")
            >>> tuplet.insert(1, second_tuplet)
            >>> show(tuplet) # doctest: +SKIP

        ..  doctest::

            >>> print(format(tuplet))
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                c'8
                \times 4/7 {
                    g'4. (
                    a'16 )
                }
                d'8
                e'8
            }


    ..  container:: example

        **Example 3.** A doubly nested tuplet:

            >>> third_tuplet = Tuplet((4, 5), [])
            >>> third_tuplet.extend("e''32 [ ef''32 d''32 cs''32 cqs''32 ]")
            >>> second_tuplet.insert(1, third_tuplet)
            >>> show(tuplet) # doctest: +SKIP

        ..  doctest::

            >>> print(format(tuplet))
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                c'8
                \tweak edge-height #'(0.7 . 0)
                \times 4/7 {
                    g'4. (
                    \times 4/5 {
                        e''32 [
                        ef''32
                        d''32
                        cs''32
                        cqs''32 ]
                    }
                    a'16 )
                }
                d'8
                e'8
            }

    '''

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Containers'

    __slots__ = (
        '_force_fraction',
        '_force_times_command',
        '_is_invisible',
        '_multiplier',
        '_preferred_denominator',
        '_signifier',
        )

    _is_counttime_component = True

    ### INITIALIZER ###

    def __init__(self, multiplier=None, music=None):
        Container.__init__(self, music)
        multiplier = multiplier or durationtools.Multiplier(2, 3)
        self.multiplier = multiplier
        self._force_fraction = False
        self._force_times_command = False
        self._is_invisible = False
        self._preferred_denominator = None
        self._signifier = '*'

    ### SPECIAL METHODS ###

    def __getnewargs__(self):
        '''Gets new arguments of tuplet.

        Returns tuple.
        '''
        return (self.multiplier,)

    ### PRIVATE PROPERTIES ###

    @property
    def _compact_representation(self):
        if not self:
            return '{{ {!s} }}'.format(self.multiplier)
        return '{{ {!s} {} }}'.format(self.multiplier, self._contents_summary)

    @property
    def _has_power_of_two_denominator(self):
        if self.multiplier:
            return mathtools.is_nonnegative_integer_power_of_two(
                self.multiplier.numerator)
        else:
            return True

    @property
    def _is_rest_filled(self):
        from abjad.tools import scoretools
        return all(isinstance(_, scoretools.Rest) for _ in self)

    @property
    def _multiplier_fraction_string(self):
        if self.preferred_denominator is not None:
            inverse_multiplier = durationtools.Multiplier(
                self.multiplier.denominator, self.multiplier.numerator)
            nonreduced_fraction = \
                mathtools.NonreducedFraction(inverse_multiplier)
            nonreduced_fraction = nonreduced_fraction.with_denominator(
                self.preferred_denominator)
            d, n = nonreduced_fraction.pair
        else:
            n, d = self.multiplier.numerator, self.multiplier.denominator
        return '%s/%s' % (n, d)

    @property
    def _preprolated_duration(self):
        return self.multiplied_duration

    @property
    def _ratio_string(self):
        multiplier = self.multiplier
        if multiplier is not None:
            numerator = multiplier.numerator
            denominator = multiplier.denominator
            ratio_string = '{}:{}'.format(denominator, numerator)
            return ratio_string
        else:
            return None

    @property
    def _summary(self):
        if 0 < len(self):
            return ', '.join([str(x) for x in self._music])
        else:
            return ''

    ### PRIVATE METHODS ###

    def _as_graphviz_node(self):
        from abjad.tools import documentationtools
        from abjad.tools import scoretools
        node = scoretools.Component._as_graphviz_node(self)
        node[0].extend([
            graphtools.GraphvizTableRow([
                graphtools.GraphvizTableCell(
                    label=type(self).__name__,
                    attributes={'border': 0},
                    ),
                ]),
            graphtools.GraphvizTableHorizontalRule(),
            graphtools.GraphvizTableRow([
                graphtools.GraphvizTableCell(
                    label='* {!s}'.format(self.multiplier),
                    attributes={'border': 0},
                    ),
                ]),
            ])
        return node

    def _fix(self):
        from abjad.tools import scoretools
        # find tuplet multiplier
        integer_exponent = int(math.log(self.multiplier, 2))
        leaf_multiplier = durationtools.Multiplier(2) ** integer_exponent
        # scale leaves in tuplet by power of two
        for component in self:
            if isinstance(component, scoretools.Leaf):
                old_written_duration = component.written_duration
                new_written_duration = leaf_multiplier * old_written_duration
                component._set_duration(new_written_duration)
        # adjust tuplet multiplier (for non-fixed-duration tuplets)
        if self.__class__ is Tuplet:
            numerator, denominator = leaf_multiplier.pair
            multiplier = durationtools.Multiplier(denominator, numerator)
            self.multiplier *= multiplier

    def _format_after_slot(self, bundle):
        result = []
        result.append(('grob reverts', bundle.grob_reverts))
        result.append(('commands', bundle.after.commands))
        result.append(('comments', bundle.after.comments))
        return tuple(result)

    def _format_before_slot(self, bundle):
        result = []
        result.append(('comments', bundle.before.comments))
        result.append(('commands', bundle.before.commands))
        result.append(('grob overrides', bundle.grob_overrides))
        result.append(('context settings', bundle.context_settings))
        return tuple(result)

    def _format_close_brackets_slot(self, bundle):
        result = []
        if self.multiplier:
            result.append([('self_brackets', 'close'), '}'])
        return tuple(result)

    def _format_closing_slot(self, bundle):
        result = []
        result.append(('commands', bundle.closing.commands))
        result.append(('comments', bundle.closing.comments))
        return self._format_slot_contributions_with_indent(result)

    def _format_lilypond_fraction_command_string(self):
        if self.is_invisible:
            return ''
        if 'text' in vars(override(self).tuplet_number):
            return''
        if (self.is_augmentation or
            not self._has_power_of_two_denominator or
                self.force_fraction):
            return r"\tweak text #tuplet-number::calc-fraction-text"
        return ''

    def _format_open_brackets_slot(self, bundle):
        result = []
        if self.multiplier:
            if self.is_invisible:
                contributor = (self, 'is_invisible')
                scale_durations_command_string = \
                    self._get_scale_durations_command_string()
                contributions = [scale_durations_command_string]
                result.append([contributor, contributions])
            else:
                contributor = ('self_brackets', 'open')
                if self.force_times_command or self.multiplier != 1:
                    contributions = []
                    fraction_command_string = \
                        self._format_lilypond_fraction_command_string()
                    if fraction_command_string:
                        contributions.append(fraction_command_string)
                    edge_height_tweak_string = \
                        self._get_edge_height_tweak_string()
                    if edge_height_tweak_string:
                        contributions.append(edge_height_tweak_string)
                    times_command_string = self._get_times_command_string()
                    contributions.append(times_command_string)
                else:
                    contributions = ['{']
                result.append([contributor, contributions])
        return tuple(result)

    def _format_opening_slot(self, bundle):
        result = []
        result.append(('comments', bundle.opening.comments))
        result.append(('commands', bundle.opening.commands))
        return self._format_slot_contributions_with_indent(result)

    def _get_edge_height_tweak_string(self):
        from abjad.tools import scoretools
        parentage = inspect_(self).get_parentage()
        measure = parentage.get_first(scoretools.Measure)
        if measure and measure.implicit_scaling:
            return
        duration = self._preprolated_duration
        denominator = duration.denominator
        if not mathtools.is_nonnegative_integer_power_of_two(denominator):
            return r"\tweak edge-height #'(0.7 . 0)"

    def _get_format_specification(self):
        return systemtools.FormatSpecification(
            client=self,
            repr_args_values=[self.multiplier, self._contents_summary],
            storage_format_args_values=[self.multiplier, self[:]],
            storage_format_kwargs_names=[],
            )

    def _get_lilypond_format(self):
        self._update_now(indicators=True)
        return self._format_component()

    def _get_scale_durations_command_string(self):
        multiplier = self.multiplier
        string = r"\scaleDurations #'({} . {}) {{"
        string = string.format(multiplier.numerator, multiplier.denominator)
        return string

    def _get_times_command_string(self):
        string = r'\times {} {{'.format(
            self._multiplier_fraction_string
            )
        return string

    def _scale(self, multiplier):
        from abjad.tools import scoretools
        multiplier = durationtools.Multiplier(multiplier)
        for component in self[:]:
            if isinstance(component, scoretools.Leaf):
                new_duration = multiplier * component.written_duration
                component._set_duration(new_duration)
        self._fix()

    def _simplify_redundant_tuplet(self):
        from abjad.tools import scoretools
        if not self.is_redundant:
            return
        leaves = []
        logical_ties = select(self).by_logical_tie(parentage_mask=self)
        durations = [_.get_duration() for _ in logical_ties]
        tuplet_duration = sum(durations)
        for i, logical_tie in enumerate(logical_ties):
            duration = durations[i]
            if i == len(logical_ties) - 1:
                leaf = logical_tie[-1]
            else:
                leaf = logical_tie[0]
            leaf.written_duration = duration
            leaves.append(leaf)
        self[:] = leaves
        #leaves = self[:]
        #leaf_durations = [inspect_(_).get_duration() for _ in leaves]
        #tuplet_duration = sum(leaf_durations)
        #for leaf_duration, leaf in zip(leaf_durations, leaves):
        #    leaf.written_duration = leaf_duration
        if isinstance(self, scoretools.FixedDurationTuplet):
            self.target_duration = tuplet_duration
        else:
            self.multiplier = durationtools.Multiplier(1)

    ### PUBLIC PROPERTIES ###

    @property
    def force_fraction(self):
        r'''Gets and sets flag to force fraction formatting of tuplet.

        ..  container:: example

            **Example 1.** Gets forced fraction formatting of tuplet:

            ::

                >>> tuplet = Tuplet((2, 3), "c'8 d'8 e'8")
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \times 2/3 {
                    c'8
                    d'8
                    e'8
                }


            ::

                >>> tuplet.force_fraction
                False

        ..  container:: example

            **Example 2.** Sets forced fraction formatting of tuplet:

            ::

                >>> tuplet.force_fraction = True
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \tweak text #tuplet-number::calc-fraction-text
                \times 2/3 {
                    c'8
                    d'8
                    e'8
                }

        ..  container:: example

            **Example 3.** Ignored when tuplet number text is overridden
            explicitly:

            ::

                >>> tuplet = Tuplet((2, 3), "c'8 d'8 e'8")
                >>> duration = inspect_(tuplet).get_duration()
                >>> markup = duration.to_score_markup()
                >>> override(tuplet).tuplet_number.text = markup
                >>> staff = Staff([tuplet])
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \override TupletNumber.text = \markup {
                        \score
                            {
                                \new Score \with {
                                    \override SpacingSpanner.spacing-increment = #0.5
                                    proportionalNotationDuration = ##f
                                } <<
                                    \new RhythmicStaff \with {
                                        \remove Time_signature_engraver
                                        \remove Staff_symbol_engraver
                                        \override Stem.direction = #up
                                        \override Stem.length = #5
                                        \override TupletBracket.bracket-visibility = ##t
                                        \override TupletBracket.direction = #up
                                        \override TupletBracket.padding = #1.25
                                        \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                        \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                        tupletFullLength = ##t
                                    } {
                                        c'4
                                    }
                                >>
                                \layout {
                                    indent = #0
                                    ragged-right = ##t
                                }
                            }
                        }
                    \times 2/3 {
                        c'8
                        d'8
                        e'8
                    }
                    \revert TupletNumber.text
                }

        Returns boolean or none.
        '''
        return self._force_fraction

    @force_fraction.setter
    def force_fraction(self, arg):
        if isinstance(arg, (bool)):
            self._force_fraction = arg
        else:
            message = 'must be true or false: {!r}.'
            message = message.format(arg)
            raise TypeError(message)

    @property
    def force_times_command(self):
        r'''Is true when trivial tuplets print LilyPond ``\times`` command.
        Otherwise false.

        ..  container:: example

            **Example 1.** Trivial tuplets normally print as a LilyPond
            container enclosed in ``{`` and ``}`` but without the LilyPond
            ``\times`` command:

            ::

                >>> trivial_tuplet = Tuplet((1, 1), "c'4 d' e'")
                >>> trivial_tuplet.force_times_command
                False

            ::

                >>> f(trivial_tuplet)
                {
                    c'4
                    d'4
                    e'4
                }

            ::

                >>> show(trivial_tuplet) # doctest: +SKIP

        ..  container:: example

            **Example 2.** But it is possible to force a trivial tuplet to
            format the LilyPond ``\times`` command:

            ::

                >>> trivial_tuplet = Tuplet((1, 1), "c'4 d' e'")
                >>> trivial_tuplet.force_times_command = True

            ::

                >>> f(trivial_tuplet)
                \times 1/1 {
                    c'4
                    d'4
                    e'4
                }


                >>> show(trivial_tuplet) # doctest: +SKIP

        ..  container:: example

            **Example 3.** This makes it possible to override tuplet number
            text:

            ::

                >>> trivial_tuplet = Tuplet((1, 1), "c'4 d' e'")
                >>> trivial_tuplet.force_times_command = True
                >>> duration = inspect_(trivial_tuplet).get_duration()
                >>> markup = duration.to_score_markup()
                >>> markup = markup.scale((0.75, 0.75))
                >>> override(trivial_tuplet).tuplet_number.text = markup
                >>> staff = Staff([trivial_tuplet])

            ::

                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> f(staff)
                \new Staff {
                    \override TupletNumber.text = \markup {
                        \scale
                            #'(0.75 . 0.75)
                            \score
                                {
                                    \new Score \with {
                                        \override SpacingSpanner.spacing-increment = #0.5
                                        proportionalNotationDuration = ##f
                                    } <<
                                        \new RhythmicStaff \with {
                                            \remove Time_signature_engraver
                                            \remove Staff_symbol_engraver
                                            \override Stem.direction = #up
                                            \override Stem.length = #5
                                            \override TupletBracket.bracket-visibility = ##t
                                            \override TupletBracket.direction = #up
                                            \override TupletBracket.padding = #1.25
                                            \override TupletBracket.shorten-pair = #'(-1 . -1.5)
                                            \override TupletNumber.text = #tuplet-number::calc-fraction-text
                                            tupletFullLength = ##t
                                        } {
                                            c'2.
                                        }
                                    >>
                                    \layout {
                                        indent = #0
                                        ragged-right = ##t
                                    }
                                }
                        }
                    \times 1/1 {
                        c'4
                        d'4
                        e'4
                    }
                    \revert TupletNumber.text
                }

        Defaults to false.

        Set to true or false.

        Returns true or false.
        '''
        return self._force_times_command

    @force_times_command.setter
    def force_times_command(self, arg):
        if isinstance(arg, (bool, type(None))):
            self._force_times_command = arg
        else:
            message = 'must be true or false: {!r}.'
            message = message.format(arg)
            raise TypeError(message)

    @property
    def implied_prolation(self):
        r'''Gets implied prolation of tuplet.

        ..  container:: example

            ::

                >>> tuplet = Tuplet((2, 3), "c'8 d'8 e'8")
                >>> show(tuplet) # doctest: +SKIP

            ::

                >>> tuplet.implied_prolation
                Multiplier(2, 3)

        Defined equal to tuplet multiplier.

        Returns multiplier.
        '''
        return self.multiplier

    @property
    def is_augmentation(self):
        r'''Is true when tuplet multiplier is greater than ``1``.
        Otherwise false.

        ..  container:: example

            **Example 1.** Augmented tuplet:

            ::

                >>> tuplet = Tuplet((4, 3), "c'8 d'8 e'8")
                >>> show(tuplet) # doctest: +SKIP

            ::

                >>> tuplet.is_augmentation
                True

        ..  container:: example

            **Example 2.** Diminished tuplet:

            ::

                >>> tuplet = Tuplet((2, 3), "c'4 d'4 e'4")
                >>> show(tuplet) # doctest: +SKIP

            ::

                >>> tuplet.is_augmentation
                False

        ..  container:: example

            **Example 3.** Trivial tuplet:

            ::

                >>> tuplet = Tuplet((1, 1), "c'8. d'8. e'8.")
                >>> show(tuplet) # doctest: +SKIP

            ::

                >>> tuplet.is_augmentation
                False

        Returns true or false.
        '''
        if self.multiplier:
            return 1 < self.multiplier
        else:
            return False

    @property
    def is_diminution(self):
        r'''Is true when tuplet multiplier is less than ``1``.
        Otherwise false.

        ..  container:: example

            **Example 1.** Augmented tuplet:

            ::

                >>> tuplet = Tuplet((4, 3), "c'8 d'8 e'8")
                >>> show(tuplet) # doctest: +SKIP

            ::

                >>> tuplet.is_diminution
                False

        ..  container:: example

            **Example 2.** Diminished tuplet:

            ::

                >>> tuplet = Tuplet((2, 3), "c'4 d'4 e'4")
                >>> show(tuplet) # doctest: +SKIP

            ::

                >>> tuplet.is_diminution
                True

        ..  container:: example

            **Example 3.** Trivial tuplet:

            ::

                >>> tuplet = Tuplet((1, 1), "c'8. d'8. e'8.")
                >>> show(tuplet) # doctest: +SKIP

            ::

                >>> tuplet.is_diminution
                False

        Returns true or false.
        '''
        if self.multiplier:
            return self.multiplier < 1
        else:
            return False

    @property
    def is_invisible(self):
        r'''Gets and sets invisibility status of tuplet.

        ..  container:: example

            **Example 1.** Gets tuplet invisibility flag:

            ::

                >>> tuplet = Tuplet((2, 3), "c'8 d'8 e'8")
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \times 2/3 {
                    c'8
                    d'8
                    e'8
                }

            ::

                >>> tuplet.is_invisible
                False

        ..  container:: example

            **Example 2.** Sets tuplet invisibility flag:

            ::

                >>> tuplet_1 = Tuplet((2, 3), "c'4 d'4 e'4")
                >>> tuplet_2 = Tuplet((2, 3), "d'4 e'4 f'4")
                >>> staff = Staff([tuplet_1, tuplet_2])
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \times 2/3 {
                        c'4
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        d'4
                        e'4
                        f'4
                    }
                }

            ::

                >>> staff[0].is_invisible = True
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \scaleDurations #'(2 . 3) {
                        c'4
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        d'4
                        e'4
                        f'4
                    }
                }

        Hides tuplet bracket and tuplet number when true.

        Preserves tuplet duration when true.

        Returns boolean or none.
        '''
        return self._is_invisible

    @is_invisible.setter
    def is_invisible(self, arg):
        assert isinstance(arg, bool), repr(arg)
        self._is_invisible = arg

    @property
    def is_redundant(self):
        r'''Is true when tuplet is redundant. Otherwise false.

        Two conditions must be true for Abjad to identify a tuplet as
        redundant. First, the tuplet must contain only leaves (not other
        tuplets). Second, the durations of all leaves contained in the tuplet
        must be able to be rewritten without a tuplet bracket.


        ..  container:: example

            **Example 1.** Redudant tuplet:

            ::

                >>> tuplet = Tuplet(Multiplier(3, 4), "c'4 c'4")
                >>> measure = Measure((3, 8), [tuplet])
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/8
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 3/4 {
                        c'4
                        c'4
                    }
                }

            ::

                >>> tuplet.is_redundant
                True

            Can be rewritten without a tuplet bracket:

                >>> measure = Measure((3, 8), "c'8. c'8.")
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/8
                    c'8.
                    c'8.
                }

        ..  container:: example

            **Example 2.** Nonredundant tuplet:

            ::

                >>> tuplet = Tuplet(Multiplier(3, 5), "c'4 c'4 c'4 c'4 c'4")
                >>> measure = Measure((3, 4), [tuplet])
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/4
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 3/5 {
                        c'4
                        c'4
                        c'4
                        c'4
                        c'4
                    }
                }

            ::

                >>> tuplet.is_redundant
                False

            Can not be rewritten without a tuplet bracket.

        Returns true or false.
        '''
        logical_ties = iterate(self).by_logical_tie(parentage_mask=self)
        return all(_.get_duration().is_assignable for _ in logical_ties)

    @property
    def is_trivial(self):
        r'''Is true when tuplet multiplier is equal to ``1``.
        Otherwise false:

        ..  container:: example

            ::

                >>> tuplet = Tuplet((1, 1), "c'8 d'8 e'8")

            ::

                >>> show(tuplet) # doctest: +SKIP

            ::

                >>> tuplet.is_trivial
                True

        Returns true or false.
        '''
        return self.multiplier == 1

    @property
    def multiplied_duration(self):
        r'''Multiplied duration of tuplet.

        ..  container:: example

            ::

                >>> tuplet = Tuplet((2, 3), "c'8 d'8 e'8")

            ::

                >>> show(tuplet) # doctest: +SKIP

            ::

                >>> tuplet.multiplied_duration
                Duration(1, 4)

        Returns duration.
        '''
        return self.multiplier * self._contents_duration

    @property
    def multiplier(self):
        r'''Gets and sets multiplier of tuplet.

        ..  container:: example

            **Example 1.** Gets tuplet multiplier:

                >>> tuplet = Tuplet((2, 3), "c'8 d'8 e'8")
                >>> show(tuplet) # doctest: +SKIP

            ::

                >>> tuplet.multiplier
                Multiplier(2, 3)

        ..  container:: example

            **Example 2.** Sets tuplet multiplier:

                >>> tuplet.multiplier = Multiplier(4, 3)
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \tweak text #tuplet-number::calc-fraction-text
                \times 4/3 {
                    c'8
                    d'8
                    e'8
                }

        Returns multiplier.
        '''
        return self._multiplier

    @multiplier.setter
    def multiplier(self, argument):
        if isinstance(argument, (int, Fraction)):
            rational = durationtools.Multiplier(argument)
        elif isinstance(argument, tuple):
            rational = durationtools.Multiplier(argument)
        else:
            message = 'can not set tuplet multiplier: {!r}.'
            message = message.format(argument)
            raise ValueError(message)
        if 0 < rational:
            self._multiplier = rational
        else:
            message = 'tuplet multiplier must be positive: {!r}.'
            message = message.format(argument)
            raise ValueError(message)

    @property
    def preferred_denominator(self):
        r'''Gets and sets preferred denominator of tuplet.

        ..  container:: example

            **Example 1.** Gets preferred denominator of tuplet:

            ::

                >>> tuplet = Tuplet((2, 3), "c'8 d'8 e'8")
                >>> tuplet.preferred_denominator is None
                True
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \times 2/3 {
                    c'8
                    d'8
                    e'8
                }

        ..  container:: example

            **Example 2.** Sets preferred denominator of tuplet:

            ::

                >>> tuplet = Tuplet((2, 3), "c'8 d'8 e'8")
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \times 2/3 {
                    c'8
                    d'8
                    e'8
                }

            ::

                >>> tuplet.preferred_denominator = 4
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \times 4/6 {
                    c'8
                    d'8
                    e'8
                }

        Returns positive integer or none.
        '''
        return self._preferred_denominator

    @preferred_denominator.setter
    def preferred_denominator(self, arg):
        if isinstance(arg, int):
            if not 0 < arg:
                raise ValueError(arg)
        elif not isinstance(arg, type(None)):
            raise TypeError(arg)
        self._preferred_denominator = arg

    ### PUBLIC METHODS ###

    @staticmethod
    def from_duration_and_ratio(
        duration,
        ratio,
        avoid_dots=True,
        decrease_durations_monotonically=True,
        is_diminution=True,
        ):
        r'''Makes tuplet from `duration` and `ratio`.

        ..  container:: example

            **Example 1.** Makes augmented tuplet from `duration` and
            `ratio` and avoid dots.

            Makes tupletted leaves strictly without dots when all
            `ratio` equal ``1``:

            ::

                >>> tuplet = Tuplet.from_duration_and_ratio(
                ...     Duration(3, 16),
                ...     mathtools.Ratio((1, 1, 1, -1, -1)),
                ...     avoid_dots=True,
                ...     is_diminution=False,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/5 {
                        c'32
                        c'32
                        c'32
                        r32
                        r32
                    }
                }

            Allows tupletted leaves to return with dots when some `ratio`
            do not equal ``1``:

            ::

                >>> tuplet = Tuplet.from_duration_and_ratio(
                ...     Duration(3, 16),
                ...     mathtools.Ratio((1, -2, -2, 3, 3)),
                ...     avoid_dots=True,
                ...     is_diminution=False,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 12/11 {
                        c'64
                        r32
                        r32
                        c'32.
                        c'32.
                    }
                }

            Interprets nonassignable `ratio` according to
            `decrease_durations_monotonically`:

            ::

                >>> tuplet = Tuplet.from_duration_and_ratio(
                ...     Duration(3, 16),
                ...     mathtools.Ratio((5, -1, 5)),
                ...     avoid_dots=True,
                ...     decrease_durations_monotonically=False,
                ...     is_diminution=False,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 12/11 {
                        c'64 ~
                        c'16
                        r64
                        c'64 ~
                        c'16
                    }
                }

        ..  container:: example

            **Example 2.** Makes augmented tuplet from `duration` and
            `ratio` and encourages dots:

            ::

                >>> tuplet = Tuplet.from_duration_and_ratio(
                ...     Duration(3, 16),
                ...     mathtools.Ratio((1, 1, 1, -1, -1)),
                ...     avoid_dots=False,
                ...     is_diminution=False,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 8/5 {
                        c'64.
                        c'64.
                        c'64.
                        r64.
                        r64.
                    }
                }

            Interprets nonassignable `ratio` according to
            `decrease_durations_monotonically`:

            ::

                >>> tuplet = Tuplet.from_duration_and_ratio(
                ...     Duration(3, 16),
                ...     mathtools.Ratio((5, -1, 5)),
                ...     avoid_dots=False,
                ...     decrease_durations_monotonically=False,
                ...     is_diminution=False,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 16/11 {
                        c'32...
                        r128.
                        c'32...
                    }
                }

        ..  container:: example

            **Example 3.** Makes diminished tuplet from `duration` and nonzero
            integer `ratio`.

            Makes tupletted leaves strictly without dots when all
            `ratio` equal ``1``:

            ::

                >>> tuplet = Tuplet.from_duration_and_ratio(
                ...     Duration(3, 16),
                ...     mathtools.Ratio((1, 1, 1, -1, -1)),
                ...     avoid_dots=True,
                ...     is_diminution=True,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 3/5 {
                        c'16
                        c'16
                        c'16
                        r16
                        r16
                    }
                }

            Allows tupletted leaves to return with dots when some `ratio`
            do not equal ``1``:

            ::

                >>> tuplet = Tuplet.from_duration_and_ratio(
                ...     Duration(3, 16),
                ...     mathtools.Ratio((1, -2, -2, 3, 3)),
                ...     avoid_dots=True,
                ...     is_diminution=True,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/11 {
                        c'32
                        r16
                        r16
                        c'16.
                        c'16.
                    }
                }

            Interprets nonassignable `ratio` according to
            `decrease_durations_monotonically`:

            ::

                >>> tuplet = Tuplet.from_duration_and_ratio(
                ...     Duration(3, 16),
                ...     mathtools.Ratio((5, -1, 5)),
                ...     avoid_dots=True,
                ...     decrease_durations_monotonically=False,
                ...     is_diminution=True,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/11 {
                        c'32 ~
                        c'8
                        r32
                        c'32 ~
                        c'8
                    }
                }

        ..  container:: example

            **Example 4.** Makes diminished tuplet from `duration` and
            `ratio` and encourages dots:

            ::

                >>> tuplet = Tuplet.from_duration_and_ratio(
                ...     Duration(3, 16),
                ...     mathtools.Ratio((1, 1, 1, -1, -1)),
                ...     avoid_dots=False,
                ...     is_diminution=True,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/16
                    \times 4/5 {
                        c'32.
                        c'32.
                        c'32.
                        r32.
                        r32.
                    }
                }

            Interprets nonassignable `ratio` according to `direction`:

            ::

                >>> tuplet = Tuplet.from_duration_and_ratio(
                ...     Duration(3, 16),
                ...     mathtools.Ratio((5, -1, 5)),
                ...     avoid_dots=False,
                ...     decrease_durations_monotonically=False,
                ...     is_diminution=True,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/16
                    \times 8/11 {
                        c'16...
                        r64.
                        c'16...
                    }
                }

        Reduces `ratio` relative to each other.

        Interprets negative `ratio` as rests.

        Returns fixed-duration tuplet.
        '''
        from abjad.tools import scoretools
        # coerce duration and ratio
        duration = durationtools.Duration(duration)
        ratio = mathtools.Ratio(ratio)
        # find basic duration of note in tuplet
        basic_prolated_duration = duration / mathtools.weight(ratio.numbers)
        # find basic written duration of note in tuplet
        if avoid_dots:
            basic_written_duration = \
                basic_prolated_duration.equal_or_greater_power_of_two
        else:
            basic_written_duration = \
                basic_prolated_duration.equal_or_greater_assignable
        # find written duration of each note in tuplet
        written_durations = [x * basic_written_duration for x in ratio.numbers]
        # make tuplet leaves
        try:
            notes = [
                scoretools.Note(0, x) if 0 < x else scoretools.Rest(abs(x))
                for x in written_durations
                ]
        except AssignabilityError:
            # changed in hopes of removing Python 3.5 ACCELERATED=True error
            #denominator = duration._denominator
            denominator = duration.denominator
            note_durations = [
                durationtools.Duration(x, denominator)
                for x in ratio.numbers
                ]
            pitches = [None if note_duration < 0 else 0
                for note_duration in note_durations]
            leaf_durations = [abs(note_duration)
                for note_duration in note_durations]
            notes = scoretools.make_leaves(
                pitches,
                leaf_durations,
                decrease_durations_monotonically=decrease_durations_monotonically,
                )
        # make tuplet
        tuplet = scoretools.FixedDurationTuplet(duration, notes)
        # fix tuplet contents if necessary
        tuplet._fix()
        # change prolation if necessary
        if not tuplet.multiplier == 1:
            if is_diminution:
                if not tuplet.is_diminution:
                    tuplet.toggle_prolation()
            else:
                if tuplet.is_diminution:
                    tuplet.toggle_prolation()
        # return tuplet
        return tuplet

    @staticmethod
    def from_leaf_and_ratio(leaf, ratio, is_diminution=True):
        r'''Makes tuplet from `leaf` and `ratio`.

        ::

            >>> note = Note("c'8.")

        ..  container:: example

            **Example 1a.** Changes leaf to augmented tuplets with `ratio`:

            ::

                >>> tuplet = Tuplet.from_leaf_and_ratio(
                ...     note,
                ...     mathtools.Ratio((1,)),
                ...     is_diminution=False,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                {
                    c'8.
                }

        ..  container:: example

            **Example 1b.** Changes leaf to augmented tuplets with `ratio`:

            ::

                >>> tuplet = Tuplet.from_leaf_and_ratio(
                ...     note,
                ...     [1, 2],
                ...     is_diminution=False,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                {
                    c'16
                    c'8
                }

        ..  container:: example

            **Example 1c.** Changes leaf to augmented tuplets with `ratio`:

            ::

                >>> tuplet = Tuplet.from_leaf_and_ratio(
                ...     note,
                ...     mathtools.Ratio((1, 2, 2)),
                ...     is_diminution=False,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \tweak text #tuplet-number::calc-fraction-text
                \times 8/5 {
                    c'64.
                    c'32.
                    c'32.
                }

        ..  container:: example

            **Example 1d.** Changes leaf to augmented tuplets with `ratio`:

            ::

                >>> tuplet = Tuplet.from_leaf_and_ratio(
                ...     note,
                ...     [1, 2, 2, 3],
                ...     is_diminution=False,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/2 {
                    c'64
                    c'32
                    c'32
                    c'32.
                }

        ..  container:: example

            **Example 1e.** Changes leaf to augmented tuplets with `ratio`:

            ::

                >>> tuplet = Tuplet.from_leaf_and_ratio(
                ...     note,
                ...     [1, 2, 2, 3, 3],
                ...     is_diminution=False,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \tweak text #tuplet-number::calc-fraction-text
                \times 12/11 {
                    c'64
                    c'32
                    c'32
                    c'32.
                    c'32.
                }

        ..  container:: example

            **Example 1f.** Changes leaf to augmented tuplets with `ratio`:

            ::

                >>> tuplet = Tuplet.from_leaf_and_ratio(
                ...     note,
                ...     mathtools.Ratio((1, 2, 2, 3, 3, 4)),
                ...     is_diminution=False,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \tweak text #tuplet-number::calc-fraction-text
                \times 8/5 {
                    c'128
                    c'64
                    c'64
                    c'64.
                    c'64.
                    c'32
                }

        ..  container:: example

            **Example 2a.** Changes leaf to diminished tuplets with `ratio`:

            ::

                >>> tuplet = Tuplet.from_leaf_and_ratio(
                ...     note,
                ...     [1],
                ...     is_diminution=True,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                {
                    c'8.
                }

        ..  container:: example

            **Example 2b.** Changes leaf to diminished tuplets with `ratio`:

            ::

                >>> tuplet = Tuplet.from_leaf_and_ratio(
                ...     note,
                ...     [1, 2],
                ...     is_diminution=True,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                {
                    c'16
                    c'8
                }

        ..  container:: example

            **Example 2c.** Changes leaf to diminished tuplets with `ratio`:

            ::

                >>> tuplet = Tuplet.from_leaf_and_ratio(
                ...     note,
                ...     [1, 2, 2],
                ...     is_diminution=True,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \times 4/5 {
                    c'32.
                    c'16.
                    c'16.
                }

        ..  container:: example

            **Example 2d.** Changes leaf to diminished tuplets with `ratio`:

            ::

                >>> tuplet = Tuplet.from_leaf_and_ratio(
                ...     note,
                ...     [1, 2, 2, 3],
                ...     is_diminution=True,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/4 {
                    c'32
                    c'16
                    c'16
                    c'16.
                }

        ..  container:: example

            **Example 2e.** Changes leaf to diminished tuplets with `ratio`:

            ::

                >>> tuplet = Tuplet.from_leaf_and_ratio(
                ...     note,
                ...     [1, 2, 2, 3, 3],
                ...     is_diminution=True,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \tweak text #tuplet-number::calc-fraction-text
                \times 6/11 {
                    c'32
                    c'16
                    c'16
                    c'16.
                    c'16.
                }

        ..  container:: example

            **Example 2f.** Changes leaf to diminished tuplets
            with `ratio`:

            ::

                >>> tuplet = Tuplet.from_leaf_and_ratio(
                ...     note,
                ...     [1, 2, 2, 3, 3, 4],
                ...     is_diminution=True,
                ...     )
                >>> measure = Measure((3, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \times 4/5 {
                    c'64
                    c'32
                    c'32
                    c'32.
                    c'32.
                    c'16
                }

        Returns tuplet.
        '''
        tuplet = leaf._to_tuplet_with_ratio(
            ratio,
            is_diminution=is_diminution,
            )
        return tuplet

    @staticmethod
    def from_nonreduced_ratio_and_nonreduced_fraction(ratio, fraction):
        r'''Makes tuplet from nonreduced `ratio` and
        nonreduced `fraction`.

        ..  container:: example

            **Example 1.** Makes container when no prolation is necessary:

            ::

                >>> tuplet = Tuplet.from_nonreduced_ratio_and_nonreduced_fraction(
                ...     mathtools.NonreducedRatio((1,)),
                ...     mathtools.NonreducedFraction(7, 16),
                ...     )
                >>> measure = Measure((7, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 7/16
                    {
                        c'4..
                    }
                }

        ..  container:: example

            **Example 2.** Makes fixed-duration tuplet when
            prolation is necessary:

            ::

                >>> tuplet = Tuplet.from_nonreduced_ratio_and_nonreduced_fraction(
                ...     mathtools.NonreducedRatio((1, 2)),
                ...     mathtools.NonreducedFraction(7, 16),
                ...     )
                >>> measure = Measure((7, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 7/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 7/6 {
                        c'8
                        c'4
                    }
                }

            ::

                >>> tuplet = Tuplet.from_nonreduced_ratio_and_nonreduced_fraction(
                ...     mathtools.NonreducedRatio((1, 2, 4)),
                ...     mathtools.NonreducedFraction(7, 16),
                ...     )
                >>> measure = Measure((7, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 7/16
                    {
                        c'16
                        c'8
                        c'4
                    }
                }

            ::

                >>> tuplet = Tuplet.from_nonreduced_ratio_and_nonreduced_fraction(
                ...     mathtools.NonreducedRatio((1, 2, 4, 1)),
                ...     mathtools.NonreducedFraction(7, 16),
                ...     )
                >>> measure = Measure((7, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 7/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 7/8 {
                        c'16
                        c'8
                        c'4
                        c'16
                    }
                }

            ::

                >>> tuplet = Tuplet.from_nonreduced_ratio_and_nonreduced_fraction(
                ...     mathtools.NonreducedRatio((1, 2, 4, 1, 2)),
                ...     mathtools.NonreducedFraction(7, 16),
                ...     )
                >>> measure = Measure((7, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 7/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 7/10 {
                        c'16
                        c'8
                        c'4
                        c'16
                        c'8
                    }
                }

            ::

                >>> tuplet = Tuplet.from_nonreduced_ratio_and_nonreduced_fraction(
                ...     mathtools.NonreducedRatio((1, 2, 4, 1, 2, 4)),
                ...     mathtools.NonreducedFraction(7, 16),
                ...     )
                >>> measure = Measure((7, 16), [tuplet])
                >>> staff = Staff([measure])
                >>> staff.context_name = 'RhythmicStaff'
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 7/16
                    \times 1/2 {
                        c'16
                        c'8
                        c'4
                        c'16
                        c'8
                        c'4
                    }
                }

        Interprets `d` as tuplet denominator.

        Returns tuplet or container.
        '''
        from abjad.tools import scoretools
        ratio = mathtools.NonreducedRatio(ratio)
        if isinstance(fraction, tuple):
            fraction = mathtools.NonreducedFraction(*fraction)
        n = fraction.numerator
        d = fraction.denominator
        duration = durationtools.Duration(fraction)
        if len(ratio.numbers) == 1:
            if 0 < ratio.numbers[0]:
                try:
                    note = scoretools.Note(0, duration)
                    return scoretools.Container([note])
                except AssignabilityError:
                    notes = scoretools.make_notes(0, duration)
                    return scoretools.Container(notes)
            elif ratio.numbers[0] < 0:
                try:
                    rest = scoretools.Rest(duration)
                    return scoretools.Container([rest])
                except AssignabilityError:
                    rests = scoretools.make_rests(duration)
                    return scoretools.Container(rests)
            else:
                message = 'no divide zer values.'
                raise ValueError(message)
        if 1 < len(ratio.numbers):
            exponent = int(
                math.log(mathtools.weight(ratio.numbers), 2) - math.log(n, 2))
            denominator = int(d * 2 ** exponent)
            music = []
            for x in ratio.numbers:
                if not x:
                    message = 'no divide zero values.'
                    raise ValueError(message)
                if 0 < x:
                    try:
                        note = scoretools.Note(0, (x, denominator))
                        music.append(note)
                    except AssignabilityError:
                        notes = scoretools.make_notes(0, (x, denominator))
                        music.extend(notes)
                else:
                    rests = scoretools.Rest((-x, denominator))
                    music.append(rests)
            return scoretools.FixedDurationTuplet(duration, music)

    def set_minimum_denominator(self, denominator):
        r'''Sets preferred denominator of tuplet to at least `denominator`.

        ..  container:: example

            **Example 1.** Sets preferred denominator of tuplet to ``8`` at
            least:

            ::

                >>> tuplet = Tuplet((3, 5), "c'4 d'8 e'8 f'4 g'2")
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \tweak text #tuplet-number::calc-fraction-text
                \times 3/5 {
                    c'4
                    d'8
                    e'8
                    f'4
                    g'2
                }

            ::

                >>> tuplet.set_minimum_denominator(8)
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \tweak text #tuplet-number::calc-fraction-text
                \times 6/10 {
                    c'4
                    d'8
                    e'8
                    f'4
                    g'2
                }

        Returns none.
        '''
        assert mathtools.is_nonnegative_integer_power_of_two(denominator)
        Duration = durationtools.Duration
        self.force_fraction = True
        durations = [
            self._contents_duration,
            self._preprolated_duration,
            Duration(1, denominator),
            ]
        nonreduced_fractions = Duration.durations_to_nonreduced_fractions(
            durations)
        self.preferred_denominator = nonreduced_fractions[1].numerator

    def to_fixed_duration_tuplet(self):
        r'''Changes tuplet to fixed-duration tuplet.

        ..  container:: example

            **Example 1.** Changes tuplet to fixed-duration tuplet:

            ::

                >>> tuplet = Tuplet((2, 3), "c'8 d'8 e'8")
                >>> show(tuplet) # doctest: +SKIP

            ::

                >>> tuplet
                Tuplet(Multiplier(2, 3), "c'8 d'8 e'8")

            ::

                >>> new_tuplet = tuplet.to_fixed_duration_tuplet()
                >>> show(new_tuplet) # doctest: +SKIP

            ::

                >>> new_tuplet
                FixedDurationTuplet(Duration(1, 4), "c'8 d'8 e'8")

        Returns new tuplet.
        '''
        from abjad.tools import scoretools
        target_duration = self._preprolated_duration
        new_tuplet = scoretools.FixedDurationTuplet(target_duration, [])
        mutate(self).swap(new_tuplet)
        return new_tuplet

    def toggle_prolation(self):
        '''Changes augmented tuplets to diminished;
        changes diminished tuplets to augmented.

        ..  container:: example

            **Example 1.** Changes augmented tuplet to diminished:

            ::

                >>> tuplet = Tuplet((4, 3), "c'8 d'8 e'8")
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \tweak text #tuplet-number::calc-fraction-text
                \times 4/3 {
                    c'8
                    d'8
                    e'8
                }

            ::

                >>> tuplet.toggle_prolation()
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \times 2/3 {
                    c'4
                    d'4
                    e'4
                }

            Multiplies the written duration of the leaves in tuplet
            by the least power of ``2`` necessary to diminshed tuplet.

        ..  container:: example

            **Example 2.** Changes diminished tuplet to augmented:

            ::

                >>> tuplet = Tuplet((2, 3), "c'4 d'4 e'4")
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \times 2/3 {
                    c'4
                    d'4
                    e'4
                }

            ::

                >>> tuplet.toggle_prolation()
                >>> show(tuplet) # doctest: +SKIP

            ..  doctest::

                >>> print(format(tuplet))
                \tweak text #tuplet-number::calc-fraction-text
                \times 4/3 {
                    c'8
                    d'8
                    e'8
                }

            Divides the written duration of the leaves in tuplet
            by the least power of ``2`` necessary to diminshed tuplet.

        Does not yet work with nested tuplets.

        Returns none.
        '''
        if self.is_diminution:
            while self.is_diminution:
                self.multiplier *= 2
                for leaf in iterate(self).by_leaf():
                    leaf.written_duration /= 2
        elif not self.is_diminution:
            while not self.is_diminution:
                self.multiplier /= 2
                for leaf in iterate(self).by_leaf():
                    leaf.written_duration *= 2



# -*- coding: utf-8 -*-
from abjad import *


def test_scoretools_fill_measures_with_repeated_notes_01():
    r'''Populates non-power-of-two measure with repeated notes.
    '''

    measure = Measure((5, 18), [])
    measure.implicit_scaling = True
    scoretools.fill_measures_with_repeated_notes(measure, Duration(1, 16))

    assert format(measure) == stringtools.normalize(
        r'''
        {
            \time 5/18
            \scaleDurations #'(8 . 9) {
                c'16
                c'16
                c'16
                c'16
                c'16
            }
        }
        '''
        )

    assert inspect_(measure).is_well_formed()



# -*- coding: utf-8 -*-
from abjad.tools import abctools
from abjad.tools import durationtools
from abjad.tools import sequencetools
from abjad.tools.topleveltools import iterate


class MutationAgent(abctools.AbjadObject):
    r'''Mutation agent.

    ..  container:: example

        Creates mutation agent for last two notes in staff:

        ::

            >>> staff = Staff("c'4 e'4 d'4 f'4")
            >>> show(staff) # doctest: +SKIP

        ::

            >>> mutate(staff[2:])
            MutationAgent(client=Selection([Note("d'4"), Note("f'4")]))

    '''

    ### CLASS VARIABLES ###

    __slots__ = (
        '_client',
        )

    ### INITIALIZER ###

    def __init__(self, client=None):
        self._client = client

    ### PUBLIC METHODS ###

    def copy(self, n=1, include_enclosing_containers=False):
        r'''Copies component and fractures crossing spanners.

        ..  todo:: Add examples.

        Returns new component.
        '''
        from abjad.tools import scoretools
        from abjad.tools import selectiontools
        if isinstance(self._client, scoretools.Component):
            selection = selectiontools.Selection(self._client)
        else:
            selection = self._client
        result = selection._copy(
            n=n,
            include_enclosing_containers=include_enclosing_containers,
            )
        if isinstance(self._client, scoretools.Component):
            if len(result) == 1:
                result = result[0]
        return result

    def eject_contents(self):
        r'''Ejects contents from outside-of-score container.

        ..  container:: example

            Ejects leaves from container:

            ::

                >>> container = Container("c'4 ~ c'4 d'4 ~ d'4")
                >>> show(container) # doctest: +SKIP

            ..  doctest::

                >>> f(container)
                {
                    c'4 ~
                    c'4
                    d'4 ~
                    d'4
                }

            Returns container contents as a selection with spanners preserved:

            ::

                >>> contents = mutate(container).eject_contents()
                >>> contents
                Selection([Note("c'4"), Note("c'4"), Note("d'4"), Note("d'4")])

            Container contents can be safely added to a new container:

            ::

                >>> staff = Staff(contents, context_name='RhythmicStaff')
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> f(staff)
                \new RhythmicStaff {
                    c'4 ~
                    c'4
                    d'4 ~
                    d'4
                }

            New container is well formed:

            ::

                >>> inspect_(staff).is_well_formed()
                True

            Old container is empty:

            ::

                >>> container
                Container()

        Returns container contents as selection.
        '''
        return self._client._eject_contents()

    def extract(self, scale_contents=False):
        r'''Extracts mutation client from score.

        Leaves children of mutation client in score.

        ..  container:: example

            Extract tuplets:

            ::

                >>> staff = Staff()
                >>> time_signature = TimeSignature((3, 4))
                >>> attach(time_signature, staff)
                >>> staff.append(Tuplet((3, 2), "c'4 e'4"))
                >>> staff.append(Tuplet((3, 2), "d'4 f'4"))
                >>> hairpin = spannertools.Hairpin('p < f')
                >>> selector = select().by_leaf(flatten=True)
                >>> leaves = selector(staff)
                >>> attach(hairpin, leaves)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \time 3/4
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 3/2 {
                        c'4 \< \p
                        e'4
                    }
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 3/2 {
                        d'4
                        f'4 \f
                    }
                }

            ::

                >>> empty_tuplet = mutate(staff[-1]).extract()
                >>> empty_tuplet = mutate(staff[0]).extract()
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \time 3/4
                    c'4 \< \p
                    e'4
                    d'4
                    f'4 \f
                }

        ..  container:: example

            Scales tuplet contents and then extracts tuplet:

            ::

                >>> staff = Staff()
                >>> time_signature = TimeSignature((3, 4))
                >>> attach(time_signature, staff)
                >>> staff.append(Tuplet((3, 2), "c'4 e'4"))
                >>> staff.append(Tuplet((3, 2), "d'4 f'4"))
                >>> hairpin = spannertools.Hairpin('p < f')
                >>> selector = select().by_leaf(flatten=True)
                >>> leaves = selector(staff)
                >>> attach(hairpin, leaves)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \time 3/4
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 3/2 {
                        c'4 \< \p
                        e'4
                    }
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 3/2 {
                        d'4
                        f'4 \f
                    }
                }

            ::

                >>> empty_tuplet = mutate(staff[-1]).extract(
                ...     scale_contents=True)
                >>> empty_tuplet = mutate(staff[0]).extract(
                ...     scale_contents=True)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \time 3/4
                    c'4. \< \p
                    e'4.
                    d'4.
                    f'4. \f
                }

        Returns mutation client.
        '''
        return self._client._extract(scale_contents=scale_contents)

    def fuse(self):
        r'''Fuses mutation client.

        ..  container:: example

            Fuses in-score leaves:

            ::

                >>> staff = Staff("c'8 d'8 e'8 f'8")
                >>> show(staff) # doctest: +SKIP

            ::

                >>> mutate(staff[1:]).fuse()
                [Note("d'4.")]
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    c'8
                    d'4.
                }

        ..  container:: example

            Fuses parent-contiguous fixed-duration tuplets in selection:

            ::

                >>> tuplet_1 = scoretools.FixedDurationTuplet(
                ...     Duration(2, 8), [])
                >>> tuplet_1.extend("c'8 d'8 e'8")
                >>> beam = spannertools.Beam()
                >>> attach(beam, tuplet_1[:])
                >>> duration = Duration(2, 16)
                >>> tuplet_2 = scoretools.FixedDurationTuplet(duration, [])
                >>> tuplet_2.extend("c'16 d'16 e'16")
                >>> slur = spannertools.Slur()
                >>> attach(slur, tuplet_2[:])
                >>> staff = Staff([tuplet_1, tuplet_2])
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \times 2/3 {
                        c'8 [
                        d'8
                        e'8 ]
                    }
                    \times 2/3 {
                        c'16 (
                        d'16
                        e'16 )
                    }
                }

            ::

                >>> tuplets = staff[:]
                >>> mutate(tuplets).fuse()
                FixedDurationTuplet(Duration(3, 8), "c'8 d'8 e'8 c'16 d'16 e'16")
                >>> show(staff) #doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \times 2/3 {
                        c'8 [
                        d'8
                        e'8 ]
                        c'16 (
                        d'16
                        e'16 )
                    }
                }

            Returns new tuplet.

            Fuses zero or more parent-contiguous `tuplets`.

            Allows in-score `tuplets`.

            Allows outside-of-score `tuplets`.

            All `tuplets` must carry the same multiplier.

            All `tuplets` must be of the same type.

        ..  container:: example

            Fuses in-score measures:

            ::

                >>> staff = Staff()
                >>> staff.append(Measure((1, 4), "c'8 d'8"))
                >>> staff.append(Measure((2, 8), "e'8 f'8"))
                >>> slur = spannertools.Slur()
                >>> leaves = select(staff).by_leaf()
                >>> attach(slur, leaves)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    {
                        \time 1/4
                        c'8 (
                        d'8
                    }
                    {
                        \time 2/8
                        e'8
                        f'8 )
                    }
                }

            ::

                >>> measures = staff[:]
                >>> mutate(measures).fuse()
                Measure((2, 4), "c'8 d'8 e'8 f'8")
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    {
                        \time 2/4
                        c'8 (
                        d'8
                        e'8
                        f'8 )
                    }
                }

        Returns fused mutation client.
        '''
        from abjad.tools import scoretools
        from abjad.tools import selectiontools
        if isinstance(self._client, scoretools.Component):
            selection = selectiontools.Selection(self._client)
            return selection._fuse()
        elif isinstance(self._client, selectiontools.Selection) and \
            self._client._all_are_contiguous_components_in_same_logical_voice(
            self._client):
            selection = selectiontools.Selection(self._client)
            return selection._fuse()

    def replace(self, recipients):
        r'''Replaces mutation client (and contents of mutation client)
        with `recipients`.

        ..  container:: example

            Replaces in-score tuplet (and children of tuplet) with notes.
            Functions exactly the same as container setitem:

                >>> tuplet_1 = Tuplet((2, 3), "c'4 d'4 e'4")
                >>> tuplet_2 = Tuplet((2, 3), "d'4 e'4 f'4")
                >>> staff = Staff([tuplet_1, tuplet_2])
                >>> hairpin = spannertools.Hairpin('p < f')
                >>> selector = select().by_leaf(flatten=True)
                >>> leaves = selector(staff)
                >>> attach(hairpin, leaves)
                >>> slur = spannertools.Slur()
                >>> attach(slur, leaves)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \times 2/3 {
                        c'4 \< \p (
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        d'4
                        e'4
                        f'4 \f )
                    }
                }

            ::

                >>> notes = scoretools.make_notes(
                ...     "c' d' e' f' c' d' e' f'",
                ...     Duration(1, 16),
                ...     )
                >>> mutate([tuplet_1]).replace(notes)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    c'16 \< \p (
                    d'16
                    e'16
                    f'16
                    c'16
                    d'16
                    e'16
                    f'16
                    \times 2/3 {
                        d'4
                        e'4
                        f'4 \f )
                    }
                }

            Preserves both hairpin and slur.

        Returns none.
        '''
        from abjad.tools import selectiontools
        if isinstance(self._client, selectiontools.Selection):
            donors = self._client
        else:
            donors = selectiontools.Selection(self._client)
        assert donors._all_are_contiguous_components_in_same_parent(donors)
        if not isinstance(recipients, selectiontools.Selection):
            recipients = selectiontools.Selection(recipients)
        assert recipients._all_are_contiguous_components_in_same_parent(
            recipients)
        if donors:
            parent, start, stop = \
                donors._get_parent_and_start_stop_indices()
            assert parent is not None, repr(donors)
            parent.__setitem__(slice(start, stop + 1), recipients)

    # TODO: fix bug in function that causes tied notes to become untied
    def replace_measure_contents(self, new_contents):
        r'''Replaces contents of measures in client with `new_contents`.

        ..  container:: example

            Replaces skip-filled measures with notes:

            ::

                >>> pairs = [(1, 8), (3, 16)]
                >>> measures = scoretools.make_spacer_skip_measures(pairs)
                >>> staff = Staff(measures)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    {
                        \time 1/8
                        s1 * 1/8
                    }
                    {
                        \time 3/16
                        s1 * 3/16
                    }
                }

            ::

                >>> notes = [Note("c'16"), Note("d'16"), Note("e'16"), Note("f'16")]
                >>> mutate(staff).replace_measure_contents(notes)
                [Measure((1, 8), "c'16 d'16"), Measure((3, 16), "e'16 f'16 s1 * 1/16")]
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    {
                        \time 1/8
                        c'16
                        d'16
                    }
                    {
                        \time 3/16
                        e'16
                        f'16
                        s1 * 1/16
                    }
                }

        Preserves duration of all measures.

        Skips measures that are too small.

        Pads extra space at end of measures with spacer skip.

        Raises stop iteration if not enough measures.

        Returns measures iterated.
        '''
        from abjad.tools import indicatortools
        from abjad.tools import scoretools
        from abjad.tools.topleveltools import attach
        from abjad.tools.topleveltools import detach
        # init return list
        result = []
        # get first measure and first time signature
        current_measure = scoretools.get_next_measure_from_component(
            self._client)
        result.append(current_measure)
        current_time_signature = current_measure.time_signature
        # to avoide pychecker slice assignment error
        #del(current_measure[:])
        current_measure.__delitem__(slice(0, len(current_measure)))
        # iterate new contents
        new_contents = list(new_contents)
        while new_contents:
            # find candidate duration of new element plus current measure
            current_element = new_contents[0]
            multiplier = current_time_signature.implied_prolation
            preprolated_duration = current_element._preprolated_duration
            duration = multiplier * preprolated_duration
            candidate_duration = current_measure._get_duration() + duration
            # if new element fits in current measure
            if candidate_duration <= current_time_signature.duration:
                current_element = new_contents.pop(0)
                current_measure._append_without_withdrawing_from_crossing_spanners(
                    current_element)
            # otherwise restore current measure and advance to next measure
            else:
                current_time_signature = \
                    indicatortools.TimeSignature(current_time_signature)
                detach(indicatortools.TimeSignature, current_measure)
                attach(current_time_signature, current_measure)
                scoretools.append_spacer_skips_to_underfull_measures(
                    [current_measure])
                current_measure = scoretools.get_next_measure_from_component(
                    current_measure)
                if current_measure is None:
                    raise StopIteration
                result.append(current_measure)
                current_time_signature = current_measure.time_signature
                # to avoid pychecker slice assignment error
                #del(current_measure[:])
                current_measure.__delitem__(slice(0, len(current_measure)))
        # restore last iterated measure
        current_time_signature = \
            indicatortools.TimeSignature(current_time_signature)
        detach(indicatortools.TimeSignature, current_measure)
        attach(current_time_signature, current_measure)
        scoretools.append_spacer_skips_to_underfull_measures(
            current_measure)
        # return iterated measures
        return result

    def respell_with_flats(self):
        r'''Respells named pitches in mutation client with flats.

        ..  todo:: Move to pitchtools package.

        ..  container:: example

            Respells notes in staff:

            ::

                >>> staff = Staff("c'8 cs'8 d'8 ef'8 e'8 f'8")
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    c'8
                    cs'8
                    d'8
                    ef'8
                    e'8
                    f'8
                }

            ::

                >>> mutate(staff).respell_with_flats()
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    c'8
                    df'8
                    d'8
                    ef'8
                    e'8
                    f'8
                }

        Returns none.
        '''
        from abjad.tools import scoretools
        for leaf in iterate(self._client).by_leaf():
            if isinstance(leaf, scoretools.Chord):
                for note_head in leaf.note_heads:
                    note_head.written_pitch = \
                        note_head.written_pitch.respell_with_flats()
            elif hasattr(leaf, 'written_pitch'):
                leaf.written_pitch = leaf.written_pitch.respell_with_flats()

    def respell_with_sharps(self):
        r'''Respells named pitches in mutation client with sharps.

        ..  todo:: Move to pitchtools package.

        ..  container:: example

            **Example 1.** Respells notes in staff:

            ::

                >>> staff = Staff("c'8 cs'8 d'8 ef'8 e'8 f'8")
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    c'8
                    cs'8
                    d'8
                    ef'8
                    e'8
                    f'8
                }

            ::

                >>> mutate(staff).respell_with_sharps()
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    c'8
                    cs'8
                    d'8
                    ds'8
                    e'8
                    f'8
                }

        Returns none.
        '''
        from abjad.tools import scoretools
        for leaf in iterate(self._client).by_leaf():
            if isinstance(leaf, scoretools.Chord):
                for note_head in leaf.note_heads:
                    note_head.written_pitch = \
                        note_head.written_pitch.respell_with_sharps()
            elif hasattr(leaf, 'written_pitch'):
                leaf.written_pitch = leaf.written_pitch.respell_with_sharps()

    def rewrite_meter(
        self,
        meter,
        boundary_depth=None,
        initial_offset=None,
        maximum_dot_count=None,
        rewrite_tuplets=True,
        use_messiaen_style_ties=False,
        ):
        r'''Rewrites the contents of logical ties in an expression to match
        `meter`.

        ..  container:: example

            **Example 1.** Rewrites the contents of a measure in a staff
            using the default meter for that measure's time signature:

            ::

                >>> parseable = "abj: | 2/4 c'2 ~ |"
                >>> parseable += "| 4/4 c'32 d'2.. ~ d'16 e'32 ~ |"
                >>> parseable += "| 2/4 e'2 |"
                >>> staff = Staff(parseable)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    {
                        \time 2/4
                        c'2 ~
                    }
                    {
                        \time 4/4
                        c'32
                        d'2.. ~
                        d'16
                        e'32 ~
                    }
                    {
                        \time 2/4
                        e'2
                    }
                }

            ::

                >>> meter = metertools.Meter((4, 4))
                >>> print(meter.pretty_rtm_format)
                (4/4 (
                    1/4
                    1/4
                    1/4
                    1/4))

            ::

                >>> mutate(staff[1][:]).rewrite_meter(meter)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    {
                        \time 2/4
                        c'2 ~
                    }
                    {
                        \time 4/4
                        c'32
                        d'8.. ~
                        d'2 ~
                        d'8..
                        e'32 ~
                    }
                    {
                        \time 2/4
                        e'2
                    }
                }

        ..  container:: example

            **Example 2.** Rewrites the contents of a measure in a staff
            using a custom meter:

            ::

                >>> staff = Staff(parseable)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    {
                        \time 2/4
                        c'2 ~
                    }
                    {
                        \time 4/4
                        c'32
                        d'2.. ~
                        d'16
                        e'32 ~
                    }
                    {
                        \time 2/4
                        e'2
                    }
                }

            ::

                >>> rtm = '(4/4 ((2/4 (1/4 1/4)) (2/4 (1/4 1/4))))'
                >>> meter = metertools.Meter(rtm)
                >>> print(meter.pretty_rtm_format) # doctest: +SKIP
                (4/4 (
                    (2/4 (
                        1/4
                        1/4))
                    (2/4 (
                        1/4
                        1/4))))

            ::

                >>> mutate(staff[1][:]).rewrite_meter(meter)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    {
                        \time 2/4
                        c'2 ~
                    }
                    {
                        \time 4/4
                        c'32
                        d'4... ~
                        d'4...
                        e'32 ~
                    }
                    {
                        \time 2/4
                        e'2
                    }
                }

        ..  container:: example

            **Example 3.** Limit the maximum number of dots per leaf using
            `maximum_dot_count`:

            ::

                >>> parseable = "abj: | 3/4 c'32 d'8 e'8 fs'4... |"
                >>> measure = parse(parseable)
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/4
                    c'32
                    d'8
                    e'8
                    fs'4...
                }

            Without constraining the `maximum_dot_count`:

            ::

                >>> mutate(measure[:]).rewrite_meter(measure)
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/4
                    c'32
                    d'16. ~
                    d'32
                    e'16. ~
                    e'32
                    fs'4...
                }

            Constraining the `maximum_dot_count` to `2`:

            ::

                >>> measure = parse(parseable)
                >>> mutate(measure[:]).rewrite_meter(
                ...     measure,
                ...     maximum_dot_count=2,
                ...     )
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/4
                    c'32
                    d'16. ~
                    d'32
                    e'16. ~
                    e'32
                    fs'8.. ~
                    fs'4
                }

            Constraining the `maximum_dot_count` to `1`:

            ::

                >>> measure = parse(parseable)
                >>> mutate(measure[:]).rewrite_meter(
                ...     measure,
                ...     maximum_dot_count=1,
                ...     )
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/4
                    c'32
                    d'16. ~
                    d'32
                    e'16. ~
                    e'32
                    fs'16. ~
                    fs'8 ~
                    fs'4
                }

            Constraining the `maximum_dot_count` to `0`:

            ::

                >>> measure = parse(parseable)
                >>> mutate(measure[:]).rewrite_meter(
                ...     measure,
                ...     maximum_dot_count=0,
                ...     )
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 3/4
                    c'32
                    d'16 ~
                    d'32 ~
                    d'32
                    e'16 ~
                    e'32 ~
                    e'32
                    fs'16 ~
                    fs'32 ~
                    fs'8 ~
                    fs'4
                }

        ..  container:: example

            **Example 4.** Split logical ties at different depths of the
            `Meter`, if those logical ties cross any offsets at that
            depth, but do not also both begin and end at any of those offsets.

            Consider the default meter for `9/8`:

            ::

                >>> meter = metertools.Meter((9, 8))
                >>> print(meter.pretty_rtm_format)
                (9/8 (
                    (3/8 (
                        1/8
                        1/8
                        1/8))
                    (3/8 (
                        1/8
                        1/8
                        1/8))
                    (3/8 (
                        1/8
                        1/8
                        1/8))))

            We can establish that meter without specifying
            a `boundary_depth`:

            ::

                >>> parseable = "abj: | 9/8 c'2 d'2 e'8 |"
                >>> measure = parse(parseable)
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 9/8
                    c'2
                    d'2
                    e'8
                }

            ::

                >>> mutate(measure[:]).rewrite_meter(measure)
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 9/8
                    c'2
                    d'4 ~
                    d'4
                    e'8
                }

            With a `boundary_depth` of `1`, logical ties which cross any offsets
            created by nodes with a depth of `1` in this Meter's rhythm
            tree - i.e.  `0/8`, `3/8`, `6/8` and `9/8` - which do not also
            begin and end at any of those offsets, will be split:

            ::

                >>> measure = parse(parseable)
                >>> mutate(measure[:]).rewrite_meter(
                ...     measure,
                ...     boundary_depth=1,
                ...     )
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 9/8
                    c'4. ~
                    c'8
                    d'4 ~
                    d'4
                    e'8
                }

            For this `9/8` meter, and this input notation, A `boundary_depth`
            of `2` causes no change, as all logical ties already align to
            multiples of `1/8`:

            ::

                >>> measure = parse(parseable)
                >>> mutate(measure[:]).rewrite_meter(
                ...     measure,
                ...     boundary_depth=2,
                ...     )
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 9/8
                    c'2
                    d'4 ~
                    d'4
                    e'8
                }

        ..  container:: example

            **Example 5.** Comparison of `3/4` and `6/8`, at `boundary_depths`
            of 0 and 1:

            ::

                >>> triple = "abj: | 3/4 2 4 || 3/4 4 2 || 3/4 4. 4. |"
                >>> triple += "| 3/4 2 ~ 8 8 || 3/4 8 8 ~ 2 |"
                >>> duples = "abj: | 6/8 2 4 || 6/8 4 2 || 6/8 4. 4. |"
                >>> duples += "| 6/8 2 ~ 8 8 || 6/8 8 8 ~ 2 |"
                >>> score = Score([Staff(triple), Staff(duples)])

            In order to see the different time signatures on each staff,
            we need to move some engravers from the Score context to the
            Staff context:

            ::

                >>> engravers = [
                ...     'Timing_translator',
                ...     'Time_signature_engraver',
                ...     'Default_bar_line_engraver',
                ...     ]
                >>> score.remove_commands.extend(engravers)
                >>> score[0].consists_commands.extend(engravers)
                >>> score[1].consists_commands.extend(engravers)
                >>> show(score) # doctest: +SKIP

            ..  doctest::

                >>> print(format(score))
                \new Score \with {
                    \remove Timing_translator
                    \remove Time_signature_engraver
                    \remove Default_bar_line_engraver
                } <<
                    \new Staff \with {
                        \consists Timing_translator
                        \consists Time_signature_engraver
                        \consists Default_bar_line_engraver
                    } {
                        {
                            \time 3/4
                            c'2
                            c'4
                        }
                        {
                            c'4
                            c'2
                        }
                        {
                            c'4.
                            c'4.
                        }
                        {
                            c'2 ~
                            c'8
                            c'8
                        }
                        {
                            c'8
                            c'8 ~
                            c'2
                        }
                    }
                    \new Staff \with {
                        \consists Timing_translator
                        \consists Time_signature_engraver
                        \consists Default_bar_line_engraver
                    } {
                        {
                            \time 6/8
                            c'2
                            c'4
                        }
                        {
                            c'4
                            c'2
                        }
                        {
                            c'4.
                            c'4.
                        }
                        {
                            c'2 ~
                            c'8
                            c'8
                        }
                        {
                            c'8
                            c'8 ~
                            c'2
                        }
                    }
                >>

            Here we establish a meter without specifying any boundary
            depth:

            ::

                >>> for measure in iterate(score).by_class(scoretools.Measure):
                ...     mutate(measure[:]).rewrite_meter(measure)
                ...
                >>> show(score) # doctest: +SKIP

            ..  doctest::

                >>> print(format(score))
                \new Score \with {
                    \remove Timing_translator
                    \remove Time_signature_engraver
                    \remove Default_bar_line_engraver
                } <<
                    \new Staff \with {
                        \consists Timing_translator
                        \consists Time_signature_engraver
                        \consists Default_bar_line_engraver
                    } {
                        {
                            \time 3/4
                            c'2
                            c'4
                        }
                        {
                            c'4
                            c'2
                        }
                        {
                            c'4.
                            c'4.
                        }
                        {
                            c'2 ~
                            c'8
                            c'8
                        }
                        {
                            c'8
                            c'8 ~
                            c'2
                        }
                    }
                    \new Staff \with {
                        \consists Timing_translator
                        \consists Time_signature_engraver
                        \consists Default_bar_line_engraver
                    } {
                        {
                            \time 6/8
                            c'2
                            c'4
                        }
                        {
                            c'4
                            c'2
                        }
                        {
                            c'4.
                            c'4.
                        }
                        {
                            c'4. ~
                            c'4
                            c'8
                        }
                        {
                            c'8
                            c'4 ~
                            c'4.
                        }
                    }
                >>

            Here we re-establish meter at a boundary depth of `1`:

            ::

                >>> for measure in iterate(score).by_class(scoretools.Measure):
                ...     mutate(measure[:]).rewrite_meter(
                ...         measure,
                ...         boundary_depth=1,
                ...         )
                ...
                >>> show(score) # doctest: +SKIP

            ..  doctest::

                >>> print(format(score))
                \new Score \with {
                    \remove Timing_translator
                    \remove Time_signature_engraver
                    \remove Default_bar_line_engraver
                } <<
                    \new Staff \with {
                        \consists Timing_translator
                        \consists Time_signature_engraver
                        \consists Default_bar_line_engraver
                    } {
                        {
                            \time 3/4
                            c'2
                            c'4
                        }
                        {
                            c'4
                            c'2
                        }
                        {
                            c'4 ~
                            c'8
                            c'8 ~
                            c'4
                        }
                        {
                            c'2 ~
                            c'8
                            c'8
                        }
                        {
                            c'8
                            c'8 ~
                            c'2
                        }
                    }
                    \new Staff \with {
                        \consists Timing_translator
                        \consists Time_signature_engraver
                        \consists Default_bar_line_engraver
                    } {
                        {
                            \time 6/8
                            c'4. ~
                            c'8
                            c'4
                        }
                        {
                            c'4
                            c'8 ~
                            c'4.
                        }
                        {
                            c'4.
                            c'4.
                        }
                        {
                            c'4. ~
                            c'4
                            c'8
                        }
                        {
                            c'8
                            c'4 ~
                            c'4.
                        }
                    }
                >>

            Note that the two time signatures are much more clearly
            disambiguated above.

        ..  container:: example

            **Example 6.** Establishing meter recursively in measures
            with nested tuplets:

            ::

                >>> parseable = "abj: | 4/4 c'16 ~ c'4 d'8. ~ "
                >>> parseable += "2/3 { d'8. ~ 3/5 { d'16 e'8. f'16 ~ } } "
                >>> parseable += "f'4 |"
                >>> measure = parse(parseable)
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 4/4
                    c'16 ~
                    c'4
                    d'8. ~
                    \times 2/3 {
                        d'8. ~
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            d'16
                            e'8.
                            f'16 ~
                        }
                    }
                    f'4
                }

            When establishing a meter on a selection of components
            which contain containers, like `Tuplets` or `Containers`,
            `metertools.rewrite_meter()` will recurse into
            those containers, treating them as measures whose time
            signature is derived from the preprolated preprolated_duration
            of the container's contents:

            ::

                >>> mutate(measure[:]).rewrite_meter(
                ...     measure,
                ...     boundary_depth=1,
                ...     )
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 4/4
                    c'4 ~
                    c'16
                    d'8. ~
                    \times 2/3 {
                        d'8 ~
                        d'16 ~
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            d'16
                            e'8 ~
                            e'16
                            f'16 ~
                        }
                    }
                    f'4
                }

        ..  container:: example

            **Example 7a.** Default rewrite behavior doesn't subdivide the
            first note in this measure because the first note in the measure
            starts at the beginning of a level-0 beat in meter:

            ::

                >>> measure = Measure((6, 8), "c'4.. c'16 ~ c'4")
                >>> meter = metertools.Meter((6, 8))
                >>> mutate(measure[:]).rewrite_meter(meter)
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 6/8
                    c'4..
                    c'16 ~
                    c'4
                }

            **Example 7b.** Setting boundary depth to 1 subdivides the first
            note in this measure:

            ::

                >>> measure = Measure((6, 8), "c'4.. c'16 ~ c'4")
                >>> meter = metertools.Meter((6, 8))
                >>> mutate(measure[:]).rewrite_meter(meter, boundary_depth=1)
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 6/8
                    c'4. ~
                    c'16
                    c'16 ~
                    c'4
                }

            **Example 7c.** Another way of doing this is by setting preferred
            boundary depth on the meter itself:

            ::

                >>> measure = Measure((6, 8), "c'4.. c'16 ~ c'4")
                >>> meter = metertools.Meter(
                ...     (6, 8),
                ...     preferred_boundary_depth=1,
                ...     )
                >>> mutate(measure[:]).rewrite_meter(meter)
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 6/8
                    c'4. ~
                    c'16
                    c'16 ~
                    c'4
                }

            This makes it possible to divide different meters in different
            ways.

        ..  container:: example

            **Example 8.** Uses Messiaen-style ties:

            ::

                >>> measure = Measure((4, 4), "c'4. c'4. c'4")
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 4/4
                    c'4.
                    c'4.
                    c'4
                }

            ::

                >>> meter = metertools.Meter((4, 4))
                >>> mutate(measure[:]).rewrite_meter(
                ...     meter,
                ...     boundary_depth=1,
                ...     use_messiaen_style_ties=True,
                ...     )
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 4/4
                    c'4
                    c'8 \repeatTie
                    c'8
                    c'4 \repeatTie
                    c'4
                }

        ..  container:: example

            **Example 9a.** Rewrites notes and tuplets:

            ::

                >>> measure = Measure((6, 4), [
                ...     Note("c'4."),
                ...     Tuplet((6, 7), "c'4. r16"),
                ...     Tuplet((6, 7), "r16 c'4."),
                ...     Note("c'4."),
                ...     ])
                >>> string = r"c'8 ~ c'8 ~ c'8 \times 6/7 { c'4. r16 }"
                >>> string += r" \times 6/7 { r16 c'4. } c'8 ~ c'8 ~ c'8"
                >>> measure = Measure((6, 4), string)
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 6/4
                    c'8 ~
                    c'8 ~
                    c'8
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/7 {
                        c'4.
                        r16
                    }
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/7 {
                        r16
                        c'4.
                    }
                    c'8 ~
                    c'8 ~
                    c'8
                }

            ::

                >>> meter = metertools.Meter((6, 4))
                >>> mutate(measure[:]).rewrite_meter(
                ...     meter,
                ...     boundary_depth=1,
                ...     )
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 6/4
                    c'4.
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/7 {
                        c'8. ~
                        c'8 ~
                        c'16
                        r16
                    }
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/7 {
                        r16
                        c'8 ~
                        c'4
                    }
                    c'4.
                }

            The tied note rewriting is good while the tuplet rewriting
            could use some adjustment.

            **Example 9b.** Rewrites notes but not tuplets:

            ::

                >>> measure = Measure((6, 4), [
                ...     Note("c'4."),
                ...     Tuplet((6, 7), "c'4. r16"),
                ...     Tuplet((6, 7), "r16 c'4."),
                ...     Note("c'4."),
                ...     ])
                >>> string = r"c'8 ~ c'8 ~ c'8 \times 6/7 { c'4. r16 }"
                >>> string += r" \times 6/7 { r16 c'4. } c'8 ~ c'8 ~ c'8"
                >>> measure = Measure((6, 4), string)
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 6/4
                    c'8 ~
                    c'8 ~
                    c'8
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/7 {
                        c'4.
                        r16
                    }
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/7 {
                        r16
                        c'4.
                    }
                    c'8 ~
                    c'8 ~
                    c'8
                }

            ::

                >>> meter = metertools.Meter((6, 4))
                >>> mutate(measure[:]).rewrite_meter(
                ...     meter,
                ...     boundary_depth=1,
                ...     rewrite_tuplets=False,
                ...     )
                >>> show(measure) # doctest: +SKIP

            ..  doctest::

                >>> print(format(measure))
                {
                    \time 6/4
                    c'4.
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/7 {
                        c'4.
                        r16
                    }
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/7 {
                        r16
                        c'4.
                    }
                    c'4.
                }

        Operates in place and returns none.
        '''
        from abjad.tools import scoretools
        from abjad.tools import selectiontools
        from abjad.tools.metertools._rewrite_meter import _rewrite_meter
        selection = self._client
        if isinstance(selection, scoretools.Container):
            selection = selection[:]
        assert isinstance(selection, selectiontools.Selection)
        result = _rewrite_meter(
            selection,
            meter,
            boundary_depth=boundary_depth,
            initial_offset=initial_offset,
            maximum_dot_count=maximum_dot_count,
            rewrite_tuplets=rewrite_tuplets,
            use_messiaen_style_ties=use_messiaen_style_ties,
            )
        return result

    def scale(self, multiplier):
        r'''Scales mutation client by `multiplier`.

        ..  container:: example

            **Example 1a.** Scales note duration by dot-generating multiplier:

            ::

                >>> staff = Staff("c'8 ( d'8 e'8 f'8 )")
                >>> show(staff) # doctest: +SKIP

            ::

                >>> mutate(staff[1]).scale(Multiplier(3, 2))
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    c'8 (
                    d'8.
                    e'8
                    f'8 )
                }

        ..  container:: example

            **Example 1b.** Scales nontrivial logical tie by dot-generating
            `multiplier`:

            ::

                >>> staff = Staff(r"c'8 \accent ~ c'8 d'8")
                >>> time_signature = TimeSignature((3, 8))
                >>> attach(time_signature, staff)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \time 3/8
                    c'8 -\accent ~
                    c'8
                    d'8
                }

            ::

                >>> logical_tie = inspect_(staff[0]).get_logical_tie()
                >>> logical_tie = mutate(logical_tie).scale(Multiplier(3, 2))
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \time 3/8
                    c'4. -\accent
                    d'8
                }

        ..  container:: example

            **Example 1c.** Scales container by dot-generating multiplier:

            ::

                >>> container = Container(r"c'8 ( d'8 e'8 f'8 )")
                >>> show(container) # doctest: +SKIP

            ..  doctest::

                >>> print(format(container))
                {
                    c'8 (
                    d'8
                    e'8
                    f'8 )
                }

            ::

                >>> mutate(container).scale(Multiplier(3, 2))
                >>> show(container) # doctest: +SKIP

            ..  doctest::

                >>> print(format(container))
                {
                    c'8. (
                    d'8.
                    e'8.
                    f'8. )
                }

        ..  container:: example

            **Example 2a.** Scales note by tie-generating multiplier:

            ::

                >>> staff = Staff("c'8 ( d'8 e'8 f'8 )")
                >>> show(staff) # doctest: +SKIP

            ::

                >>> mutate(staff[1]).scale(Multiplier(5, 4))
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    c'8 (
                    d'8 ~
                    d'32
                    e'8
                    f'8 )
                }

        ..  container:: example

            **Example 2b.** Scales nontrivial logical tie
            by tie-generating `multiplier`:

            ::

                >>> staff = Staff(r"c'8 \accent ~ c'8 d'16")
                >>> time_signature = TimeSignature((5, 16))
                >>> attach(time_signature, staff)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \time 5/16
                    c'8 -\accent ~
                    c'8
                    d'16
                }

            ::

                >>> logical_tie = inspect_(staff[0]).get_logical_tie()
                >>> logical_tie = mutate(logical_tie).scale(Multiplier(5, 4))
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \time 5/16
                    c'4 -\accent ~
                    c'16
                    d'16
                }

        ..  container:: example

            **Example 2c.** Scales container by tie-generating multiplier:

            ::

                >>> container = Container(r"c'8 ( d'8 e'8 f'8 )")
                >>> show(container) # doctest: +SKIP

            ..  doctest::

                >>> print(format(container))
                {
                    c'8 (
                    d'8
                    e'8
                    f'8 )
                }

            ::

                >>> mutate(container).scale(Multiplier(5, 4))
                >>> show(container) # doctest: +SKIP

            ..  doctest::

                >>> print(format(container))
                {
                    c'8 ~ (
                    c'32
                    d'8 ~
                    d'32
                    e'8 ~
                    e'32
                    f'8 ~
                    f'32 )
                }

        ..  container:: example

            **Example 3a.** Scales note by tuplet-generating multiplier:

            ::

                >>> staff = Staff("c'8 ( d'8 e'8 f'8 )")
                >>> show(staff) # doctest: +SKIP

            ::

                >>> mutate(staff[1]).scale(Multiplier(2, 3))
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    c'8 (
                    \tweak edge-height #'(0.7 . 0)
                    \times 2/3 {
                        d'8
                    }
                    e'8
                    f'8 )
                }

        ..  container:: example

            **Example 3b.** Scales trivial logical tie by tuplet-generating
            multiplier:

            ::

                >>> staff = Staff(r"c'8 \accent")
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    c'8 -\accent
                }

            ::

                >>> logical_tie = inspect_(staff[0]).get_logical_tie()
                >>> logical_tie = mutate(logical_tie).scale(Multiplier(4, 3))
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \tweak edge-height #'(0.7 . 0)
                    \times 2/3 {
                        c'4 -\accent
                    }
                }

        ..  container:: example

            **Example 3c.** Scales container by tuplet-generating multiplier:

            ::

                >>> container = Container(r"c'8 ( d'8 e'8 f'8 )")
                >>> show(container) # doctest: +SKIP

            ..  doctest::

                >>> print(format(container))
                {
                    c'8 (
                    d'8
                    e'8
                    f'8 )
                }

            ::

                >>> mutate(container).scale(Multiplier(4, 3))
                >>> show(container) # doctest: +SKIP

            ..  doctest::

                >>> print(format(container))
                {
                    \tweak edge-height #'(0.7 . 0)
                    \times 2/3 {
                        c'4 (
                    }
                    \tweak edge-height #'(0.7 . 0)
                    \times 2/3 {
                        d'4
                    }
                    \tweak edge-height #'(0.7 . 0)
                    \times 2/3 {
                        e'4
                    }
                    \tweak edge-height #'(0.7 . 0)
                    \times 2/3 {
                        f'4 )
                    }
                }

        ..  container:: example

            **Example 4.** Scales note by tie- and tuplet-generating
            multiplier:

            ::

                >>> staff = Staff("c'8 ( d'8 e'8 f'8 )")
                >>> show(staff) # doctest: +SKIP

            ::

                >>> mutate(staff[1]).scale(Multiplier(5, 6))
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    c'8 (
                    \tweak edge-height #'(0.7 . 0)
                    \times 2/3 {
                        d'8 ~
                        d'32
                    }
                    e'8
                    f'8 )
                }

        ..  container:: example

            **Example 5.** Scales note carrying LilyPond multiplier:

            ::

                >>> note = Note("c'8")
                >>> attach(Multiplier(1, 2), note)
                >>> show(note) # doctest: +SKIP

            ..  doctest::

                >>> print(format(note))
                c'8 * 1/2

            ::

                >>> mutate(note).scale(Multiplier(5, 3))
                >>> show(note) # doctest: +SKIP

            ..  doctest::

                >>> print(format(note))
                c'8 * 5/6

        ..  container:: example

            **Example 6.** Scales tuplet:

            ::

                >>> staff = Staff()
                >>> time_signature = TimeSignature((4, 8))
                >>> attach(time_signature, staff)
                >>> tuplet = scoretools.Tuplet((4, 5), [])
                >>> tuplet.extend("c'8 d'8 e'8 f'8 g'8")
                >>> staff.append(tuplet)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \time 4/8
                    \times 4/5 {
                        c'8
                        d'8
                        e'8
                        f'8
                        g'8
                    }
                }

            ::

                >>> mutate(tuplet).scale(Multiplier(2))
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \time 4/8
                    \times 4/5 {
                        c'4
                        d'4
                        e'4
                        f'4
                        g'4
                    }
                }

        ..  container:: example

            **Example 7.** Scales fixed-duration tuplet:

            ::

                >>> staff = Staff()
                >>> time_signature = TimeSignature((4, 8))
                >>> attach(time_signature, staff)
                >>> tuplet = scoretools.FixedDurationTuplet((4, 8), [])
                >>> tuplet.extend("c'8 d'8 e'8 f'8 g'8")
                >>> staff.append(tuplet)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \time 4/8
                    \times 4/5 {
                        c'8
                        d'8
                        e'8
                        f'8
                        g'8
                    }
                }

            ::

                >>> mutate(tuplet).scale(Multiplier(2))
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \time 4/8
                    \times 4/5 {
                        c'4
                        d'4
                        e'4
                        f'4
                        g'4
                    }
                }

        Returns none.
        '''
        from abjad.tools import scoretools
        from abjad.tools import selectiontools
        if hasattr(self._client, '_scale'):
            self._client._scale(multiplier)
        else:
            assert isinstance(self._client, selectiontools.Selection)
            for component in self._client:
                component._scale(multiplier)

    def splice(
        self,
        components,
        direction=Right,
        grow_spanners=True,
        ):
        r'''Splices `components` to the right or left of selection.

        ..  todo:: Add examples.

        Returns list of components.
        '''
        return self._client._splice(
            components,
            direction=direction,
            grow_spanners=grow_spanners,
            )

    # TODO: fix bug that unintentionally fractures ties.
    # TODO: add tests of tupletted notes and rests.
    # TODO: add examples that show indicator handling.
    # TODO: add example showing grace and after grace handling.
    def split(
        self,
        durations,
        fracture_spanners=False,
        cyclic=False,
        tie_split_notes=True,
        use_messiaen_style_ties=False,
        ):
        r'''Splits component or selection by `durations`.

        ..  container:: example

            Splits leaves:

            ::

                >>> staff = Staff("c'8 e' d' f' c' e' d' f'")
                >>> leaves = staff[:]
                >>> hairpin = spannertools.Hairpin(descriptor='p < f')
                >>> attach(hairpin, leaves)
                >>> override(staff).dynamic_line_spanner.staff_padding = 3
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff \with {
                    \override DynamicLineSpanner.staff-padding = #3
                } {
                    c'8 \< \p
                    e'8
                    d'8
                    f'8
                    c'8
                    e'8
                    d'8
                    f'8 \f
                }

            ::

                >>> durations = [Duration(3, 16), Duration(7, 32)]
                >>> result = mutate(leaves).split(
                ...     durations,
                ...     tie_split_notes=False,
                ...     )
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff \with {
                    \override DynamicLineSpanner.staff-padding = #3
                } {
                    c'8 \< \p
                    e'16
                    e'16
                    d'8
                    f'32
                    f'16.
                    c'8
                    e'8
                    d'8
                    f'8 \f
                }

        ..  container:: example

            Splits leaves and fracture crossing spanners:

            ::

                >>> staff = Staff("c'8 e' d' f' c' e' d' f'")
                >>> hairpin = spannertools.Hairpin(descriptor='p < f')
                >>> attach(hairpin, staff[:])
                >>> override(staff).dynamic_line_spanner.staff_padding = 3
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff \with {
                    \override DynamicLineSpanner.staff-padding = #3
                } {
                    c'8 \< \p
                    e'8
                    d'8
                    f'8
                    c'8
                    e'8
                    d'8
                    f'8 \f
                }

            ::

                >>> durations = [Duration(3, 16), Duration(7, 32)]
                >>> leaves = staff[:]
                >>> result = mutate(leaves).split(
                ...     durations,
                ...     fracture_spanners=True,
                ...     tie_split_notes=False,
                ...     )
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff \with {
                    \override DynamicLineSpanner.staff-padding = #3
                } {
                    c'8 \< \p
                    e'16 \f
                    e'16 \< \p
                    d'8
                    f'32 \f
                    f'16. \< \p
                    c'8
                    e'8
                    d'8
                    f'8 \f
                }

        ..  container:: example

            Splits leaves cyclically:

            ::

                >>> staff = Staff("c'8 e' d' f' c' e' d' f'")
                >>> leaves = staff[:]
                >>> hairpin = spannertools.Hairpin(descriptor='p < f')
                >>> attach(hairpin, leaves)
                >>> override(staff).dynamic_line_spanner.staff_padding = 3
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff \with {
                    \override DynamicLineSpanner.staff-padding = #3
                } {
                    c'8 \< \p
                    e'8
                    d'8
                    f'8
                    c'8
                    e'8
                    d'8
                    f'8 \f
                }

            ::

                >>> durations = [Duration(3, 16), Duration(7, 32)]
                >>> result = mutate(leaves).split(
                ...     durations,
                ...     cyclic=True,
                ...     tie_split_notes=False,
                ...     )
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff \with {
                    \override DynamicLineSpanner.staff-padding = #3
                } {
                    c'8 \< \p
                    e'16
                    e'16
                    d'8
                    f'32
                    f'16.
                    c'16.
                    c'32
                    e'8
                    d'16
                    d'16
                    f'8 \f
                }

        ..  container:: example

            Splits leaves cyclically and fracture spanners:

            ::

                >>> staff = Staff("c'8 e' d' f' c' e' d' f'")
                >>> leaves = staff[:]
                >>> hairpin = spannertools.Hairpin(descriptor='p < f')
                >>> attach(hairpin, leaves)
                >>> override(staff).dynamic_line_spanner.staff_padding = 3
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff \with {
                    \override DynamicLineSpanner.staff-padding = #3
                } {
                    c'8 \< \p
                    e'8
                    d'8
                    f'8
                    c'8
                    e'8
                    d'8
                    f'8 \f
                }

            ::

                >>> durations = [Duration(3, 16), Duration(7, 32)]
                >>> result = mutate(leaves).split(
                ...     durations,
                ...     cyclic=True,
                ...     fracture_spanners=True,
                ...     tie_split_notes=False,
                ...     )
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff \with {
                    \override DynamicLineSpanner.staff-padding = #3
                } {
                    c'8 \< \p
                    e'16 \f
                    e'16 \< \p
                    d'8
                    f'32 \f
                    f'16. \< \p
                    c'16. \f
                    c'32 \< \p
                    e'8
                    d'16 \f
                    d'16 \< \p
                    f'8 \f
                }

        ..  container:: example

            Splits tupletted leaves and fracture crossing spanners:

            ::

                >>> staff = Staff()
                >>> staff.append(Tuplet((2, 3), "c'4 d' e'"))
                >>> staff.append(Tuplet((2, 3), "c'4 d' e'"))
                >>> selector = select().by_leaf(flatten=True)
                >>> leaves = selector(staff)
                >>> slur = spannertools.Slur()
                >>> attach(slur, leaves)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \times 2/3 {
                        c'4 (
                        d'4
                        e'4
                    }
                    \times 2/3 {
                        c'4
                        d'4
                        e'4 )
                    }
                }

            ::

                >>> durations = [Duration(1, 4)]
                >>> result = mutate(leaves).split(
                ...     durations,
                ...     fracture_spanners=True,
                ...     tie_split_notes=False,
                ...     )
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \times 2/3 {
                        c'4 (
                        d'8 )
                        d'8 (
                        e'4
                    }
                    \times 2/3 {
                        c'4
                        d'4
                        e'4 )
                    }
                }

        ..  container:: example

            Splits leaves cyclically and ties split notes:

            ::

                >>> staff = Staff("c'1 d'1")
                >>> hairpin = spannertools.Hairpin(descriptor='p < f')
                >>> attach(hairpin, staff[:])
                >>> override(staff).dynamic_line_spanner.staff_padding = 3
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff \with {
                    \override DynamicLineSpanner.staff-padding = #3
                } {
                    c'1 \< \p
                    d'1 \f
                }

            ::

                >>> durations = [Duration(3, 4)]
                >>> result = mutate(staff[:]).split(
                ...     durations,
                ...     cyclic=True,
                ...     fracture_spanners=False,
                ...     tie_split_notes=True,
                ...     )
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff \with {
                    \override DynamicLineSpanner.staff-padding = #3
                } {
                    c'2. ~ \< \p
                    c'4
                    d'2 ~
                    d'2 \f
                }

            As above but with Messiaen-style ties:

            ::

                >>> staff = Staff("c'1 d'1")
                >>> hairpin = spannertools.Hairpin(descriptor='p < f')
                >>> attach(hairpin, staff[:])
                >>> override(staff).dynamic_line_spanner.staff_padding = 3

            ::

                >>> durations = [Duration(3, 4)]
                >>> result = mutate(staff[:]).split(
                ...     durations,
                ...     cyclic=True,
                ...     fracture_spanners=False,
                ...     tie_split_notes=True,
                ...     use_messiaen_style_ties=True,
                ...     )
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff \with {
                    \override DynamicLineSpanner.staff-padding = #3
                } {
                    c'2. \< \p
                    c'4 \repeatTie
                    d'2
                    d'2 \repeatTie \f
                }

        Returns list of selections.
        '''
        from abjad.tools import scoretools
        from abjad.tools import selectiontools
        # check input
        components = self._client
        single_component_input = False
        if isinstance(components, scoretools.Component):
            single_component_input = True
            components = selectiontools.Selection(components)
        assert all(isinstance(x, scoretools.Component) for x in components)
        if not isinstance(components, selectiontools.Selection):
            components = selectiontools.Selection(components)
        durations = [durationtools.Duration(x) for x in durations]
        # return if no split to be done
        if not durations:
            if single_component_input:
                return components
            else:
                return [], components
        # calculate total component duration
        total_component_duration = components.get_duration()
        total_split_duration = sum(durations)
        # calculate durations
        if cyclic:
            durations = sequencetools.Sequence(durations)
            durations = durations.repeat_to_weight(total_component_duration)
            durations = list(durations)
        elif total_split_duration < total_component_duration:
            final_offset = total_component_duration - sum(durations)
            durations.append(final_offset)
        elif total_component_duration < total_split_duration:
            durations = sequencetools.Sequence(durations).truncate(
                weight=total_component_duration,
                )
            durations = list(durations)
        # keep copy of durations to partition result components
        durations_copy = durations[:]
        # calculate total split duration
        total_split_duration = sum(durations)
        assert total_split_duration == total_component_duration
        # initialize loop variables
        result, shard = [], []
        offset_index, offset_count = 0, len(durations)
        current_shard_duration = durationtools.Duration(0)
        remaining_components = list(components[:])
        advance_to_next_offset = True
        # loop and build shards by grabbing next component
        # and next duration each time through loop
        while True:
            # grab next split point
            if advance_to_next_offset:
                if durations:
                    next_split_point = durations.pop(0)
                else:
                    break
            advance_to_next_offset = True
            # grab next component from input stack of components
            if remaining_components:
                current_component = remaining_components.pop(0)
            else:
                break
            # find where current component endpoint will position us
            candidate_shard_duration = current_shard_duration + \
                current_component._get_duration()
            # if current component would fill current shard exactly
            if candidate_shard_duration == next_split_point:
                shard.append(current_component)
                result.append(shard)
                shard = []
                current_shard_duration = durationtools.Duration(0)
                offset_index += 1
            # if current component would exceed current shard
            elif next_split_point < candidate_shard_duration:
                local_split_duration = \
                    next_split_point - current_shard_duration
                if isinstance(current_component, scoretools.Leaf):
                    leaf_split_durations = [local_split_duration]
                    current_duration = current_component._get_duration()
                    additional_required_duration = \
                        current_duration - local_split_duration
                    split_durations = sequencetools.Sequence(durations)
                    split_durations = split_durations.split(
                        [additional_required_duration],
                        cyclic=False,
                        overhang=True,
                        )
                    split_durations = [list(_) for _ in split_durations]
                    additional_durations = split_durations[0]
                    leaf_split_durations.extend(additional_durations)
                    durations = split_durations[-1]
                    leaf_shards = current_component._split(
                        leaf_split_durations,
                        cyclic=False,
                        fracture_spanners=fracture_spanners,
                        tie_split_notes=tie_split_notes,
                        use_messiaen_style_ties=use_messiaen_style_ties,
                        )
                    shard.extend(leaf_shards)
                    result.append(shard)
                    offset_index += len(additional_durations)
                else:
                    left_list, right_list = \
                        current_component._split_by_duration(
                        local_split_duration,
                        fracture_spanners=fracture_spanners,
                        tie_split_notes=tie_split_notes,
                        use_messiaen_style_ties=use_messiaen_style_ties,
                        )
                    shard.extend(left_list)
                    result.append(shard)
                    remaining_components.__setitem__(slice(0, 0), right_list)
                shard = []
                offset_index += 1
                current_shard_duration = durationtools.Duration(0)
            # if current component would not fill current shard
            elif candidate_shard_duration < next_split_point:
                shard.append(current_component)
                current_shard_duration += current_component._get_duration()
                advance_to_next_offset = False
            else:
                raise ValueError
        # append any stub shard
        if len(shard):
            result.append(shard)
        # append any unexamined components
        if len(remaining_components):
            result.append(remaining_components)
        # partition split components according to input durations
        result = sequencetools.Sequence(result).flatten()
        result = selectiontools.Selection(result)
        result = result.partition_by_durations(durations_copy, fill=Exact)
        # return list of shards
        result = [selectiontools.Selection(x) for x in result]
        return result

    def swap(self, container):
        r'''Swaps mutation client for empty `container`.

        ..  container:: example

            Swaps measures for tuplet:

                >>> staff = Staff()
                >>> staff.append(Measure((3, 4), "c'4 d'4 e'4"))
                >>> staff.append(Measure((3, 4), "d'4 e'4 f'4"))
                >>> selector = select().by_leaf(flatten=True)
                >>> leaves = selector(staff)
                >>> hairpin = spannertools.Hairpin('p < f')
                >>> attach(hairpin, leaves)
                >>> measures = staff[:]
                >>> slur = spannertools.Slur()
                >>> attach(slur, leaves)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    {
                        \time 3/4
                        c'4 \< \p (
                        d'4
                        e'4
                    }
                    {
                        d'4
                        e'4
                        f'4 \f )
                    }
                }

            ::

                >>> measures = staff[:]
                >>> tuplet = Tuplet(Multiplier(2, 3), [])
                >>> tuplet.preferred_denominator = 4
                >>> mutate(measures).swap(tuplet)
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    \times 4/6 {
                        c'4 \< \p (
                        d'4
                        e'4
                        d'4
                        e'4
                        f'4 \f )
                    }
                }

        Returns none.
        '''
        from abjad.tools import scoretools
        from abjad.tools import selectiontools
        Selection = selectiontools.Selection
        if isinstance(self._client, selectiontools.Selection):
            donors = self._client
        else:
            donors = selectiontools.Selection(self._client)
        assert donors._all_are_contiguous_components_in_same_parent(donors)
        assert isinstance(container, scoretools.Container)
        assert not container, repr(container)
        donors._give_music_to_empty_container(container)
        donors._give_dominant_spanners([container])
        donors._give_position_in_parent_to_container(container)

    def transpose(self, argument):
        r'''Transposes notes and chords in mutation client by `argument`.

        ..  todo:: Move to pitchtools package.

        ..  container:: example

            Transposes notes and chords in staff:

            ::

                >>> staff = Staff()
                >>> staff.append(Measure((4, 4), "c'4 d'4 e'4 r4"))
                >>> staff.append(Measure((3, 4), "d'4 e'4 <f' a' c''>4"))
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                    >>> print(format(staff))
                    \new Staff {
                        {
                            \time 4/4
                            c'4
                            d'4
                            e'4
                            r4
                        }
                        {
                            \time 3/4
                            d'4
                            e'4
                            <f' a' c''>4
                        }
                    }

            ::

                >>> mutate(staff).transpose("+m3")
                >>> show(staff) # doctest: +SKIP

            ..  doctest::

                >>> print(format(staff))
                \new Staff {
                    {
                        \time 4/4
                        ef'4
                        f'4
                        g'4
                        r4
                    }
                    {
                        \time 3/4
                        f'4
                        g'4
                        <af' c'' ef''>4
                    }
                }

        Returns none.
        '''
        from abjad.tools import pitchtools
        from abjad.tools import scoretools
        named_interval = pitchtools.NamedInterval(argument)
        for x in iterate(self._client).by_class(
            (scoretools.Note, scoretools.Chord)):
            if isinstance(x, scoretools.Note):
                old_written_pitch = x.note_head.written_pitch
                new_written_pitch = old_written_pitch.transpose(named_interval)
                x.note_head.written_pitch = new_written_pitch
            else:
                for note_head in x.note_heads:
                    old_written_pitch = note_head.written_pitch
                    new_written_pitch = old_written_pitch.transpose(
                        named_interval)
                    note_head.written_pitch = new_written_pitch

    ### PUBLIC PROPERTIES ###

    @property
    def client(self):
        r'''Returns client of mutation agent.

        Returns selection or component.
        '''
        return self._client



# -*- coding: utf-8 -*-
from abjad import *


def test_scoretools_Chord___repr___01():
    r'''Chord repr is evaluable.
    '''

    chord_1 = Chord("<ef' cs'' f''>4")
    chord_2 = eval(repr(chord_1))

    assert isinstance(chord_1, Chord)
    assert isinstance(chord_2, Chord)
    assert format(chord_1) == format(chord_2)
    assert chord_1 is not chord_2


def test_scoretools_Chord___repr___02():
    '''Chord repr works with forced and cautionary accidentals.
    '''

    chord = Chord('<c! e? g!? b>4')
    assert "Chord('{}')".format(format(chord)) == repr(chord)



# -*- coding: utf-8 -*-


def is_assignable_integer(argument):
    r'''Is true when `argument` is equivalent to an integer that can be written
    without recourse to ties. Otherwise false.

    ..  container:: example

        ::

            >>> for n in range(0, 16 + 1):
            ...     print('%s\t%s' % (n, mathtools.is_assignable_integer(n)))
            ... 
            0  False
            1  True
            2  True
            3  True
            4  True
            5  False
            6  True
            7  True
            8  True
            9  False
            10 False
            11 False
            12 True
            13 False
            14 True
            15 True
            16 True

    Returns true or false.
    '''
    from abjad.tools import mathtools
    if isinstance(argument, int):
        if 0 < argument:
            if not '01' in mathtools.integer_to_binary_string(argument):
                return True
    return False



# -*- coding: utf-8 -*-
from abjad.tools.lilypondnametools.LilyPondNameManager \
    import LilyPondNameManager


class LilyPondTweakManager(LilyPondNameManager):
    '''LilyPond tweak manager.
    '''

    ### PRIVATE METHODS ###

    def _list_format_contributions(self):
        from abjad.tools import systemtools
        manager = systemtools.LilyPondFormatManager
        result = []
        for attribute_pair in self._get_attribute_pairs():
            assert len(attribute_pair) == 2, repr(attribute_pair)
            attribute_name, attribute_value = attribute_pair
            string = manager.make_lilypond_tweak_string(
                attribute_name,
                attribute_value,
                )
            result.append(string)
        result.sort()
        return result



# -*- coding: utf-8 -*-
from abjad import *


def test_scoretools_MultimeasureRest___init___01():
    r'''Initializes multimeasure rest from empty input.
    '''

    multimeasure_rest = scoretools.MultimeasureRest()

    assert format(multimeasure_rest) == stringtools.normalize(
        r'''
        R4
        '''
        )

    assert inspect_(multimeasure_rest).is_well_formed()



# -*- coding: utf-8 -*-
from abjad import *


def test_pitchtools_PitchRange___format___01():

    pitch_range = pitchtools.PitchRange('[A0, C8]')

    assert format(pitch_range) == stringtools.normalize(
        r'''
        pitchtools.PitchRange(
            range_string='[A0, C8]',
            )
        ''',
        )



# -*- coding: utf-8 -*-
from abjad import *
from abjad.tools.lilypondparsertools import LilyPondParser


def test_lilypondparsertools_LilyPondParser__indicators__KeySignature_01():

    target = Staff([Note("fs'", 1)])
    key_signature = KeySignature('g', 'major')
    attach(key_signature, target[0])

    assert format(target) == stringtools.normalize(
        r'''
        \new Staff {
            \key g \major
            fs'1
        }
        '''
        )

    parser = LilyPondParser()
    result = parser(format(target))
    assert format(target) == format(result) and target is not result
    key_signatures = inspect_(result[0]).get_indicators(KeySignature)
    assert len(key_signatures) == 1



# -*- coding: utf-8 -*-
from abjad import *


def test_stringtools_is_snake_case_file_name_with_extension_01():

    assert stringtools.is_snake_case_file_name_with_extension('foo_bar.blah')
    assert stringtools.is_snake_case_file_name_with_extension('foo.blah')


def test_stringtools_is_snake_case_file_name_with_extension_02():

    assert stringtools.is_snake_case_file_name_with_extension('')


def test_stringtools_is_snake_case_file_name_with_extension_03():

    assert not stringtools.is_snake_case_file_name_with_extension('foo_bar')
    assert not stringtools.is_snake_case_file_name_with_extension('foo_bar.')
    assert not stringtools.is_snake_case_file_name_with_extension('Foo_bar.blah')
    assert not stringtools.is_snake_case_file_name_with_extension('foo_bar.Blah')



# -*- coding: utf-8 -*-
from __future__ import print_function
import abc
import argparse
import inspect
import os
from abjad.tools import abctools
from abjad.tools import documentationtools
from abjad.tools import stringtools
try:
    from ConfigParser import ConfigParser
except ImportError:
    from configparser import ConfigParser
try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib


class CommandlineScript(abctools.AbjadObject):
    r'''Object-oriented model of a developer script.

    `CommandlineScript` is the abstract parent from which concrete developer
    scripts inherit.

    Developer scripts can be called from the command line, generally via the
    `ajv` command.

    Developer scripts can be instantiated by other developer scripts in order
    to share functionality.
    '''

    ### CLASS VARIABLES ###

    __slots__ = (
        '_argument_parser',
        '_config_parser',
        )

    _colors = {
        'BLUE': '\033[94m',
        'RED': '\033[91m',
        'GREEN': '\033[92m',
        'END': '\033[0m',
        }

    ### INITIALIZER ###

    def __init__(self):
        short_description = getattr(self, 'short_description', None)
        long_description = getattr(self, 'long_description', None)
        if long_description:
            long_description = stringtools.normalize(long_description)
        parser = self._argument_parser = argparse.ArgumentParser(
            description=short_description,
            epilog=long_description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            prog=self.program_name,
            )
        version = '%(prog)s {}'
        version = version.format(getattr(self, 'version', 1.0))
        parser.add_argument('--version', action='version', version=version)
        self._argument_parser = parser
        self._setup_argument_parser(parser)
        self._config_parser = None

    ### SPECIAL METHODS ###

    def __call__(self, arguments=None):
        r'''Calls developer script.

        Returns none.
        '''
        self._config_parser = self._read_config_files()
        if arguments is None:
            arguments = self.argument_parser.parse_args()
        else:
            if isinstance(arguments, str):
                arguments = arguments.split()
            elif not isinstance(arguments, (list, tuple)):
                raise ValueError
            arguments = self.argument_parser.parse_args(arguments)
        self._process_args(arguments)

    def __getstate__(self):
        r'''Gets object state.
        '''
        return {}

    ### PRIVATE METHODS ###

    def _is_valid_path(self, path):
        if os.path.exists(path):
            if os.path.isdir(path):
                return True
        return False

    @abc.abstractmethod
    def _process_args(self, arguments):
        raise NotImplementedError

    def _read_config_files(self):
        paths = []
        home_config = pathlib.Path(os.path.expanduser('~/.ajv'))
        if home_config.exists():
            paths.append(home_config)
        path = pathlib.Path.cwd()
        while not path.joinpath('.ajv').exists():
            path = path.parent
            if path.parent == path:
                break
        if path.joinpath('.ajv').exists():
            if path.joinpath('.ajv') not in paths:
                paths.append(path.joinpath('.ajv'))
        paths = [str(_) for _ in paths]
        config_parser = ConfigParser()
        config_parser.read(paths)
        return config_parser

    @abc.abstractmethod
    def _setup_argument_parser(self, parser):
        raise NotImplementedError

    def _validate_path(self, path):
        message = '{!r} is not a valid directory.'
        message = message.format(path)
        error = argparse.ArgumentTypeError(message)
        path = os.path.abspath(path)
        if not self._is_valid_path(path):
            raise error
        return os.path.relpath(path)

    ### PUBLIC METHODS ###

    @staticmethod
    def list_commandline_script_classes():
        r'''Returns a list of all developer script classes.
        '''
        from abjad.tools import abjadbooktools
        from abjad.tools import commandlinetools
        tools_package_paths = []
        tools_package_paths.extend(abjadbooktools.__path__)
        tools_package_paths.extend(commandlinetools.__path__)
        script_classes = []
        base_class = commandlinetools.CommandlineScript
        for tools_package_path in tools_package_paths:
            generator = documentationtools.yield_all_classes(
                code_root=tools_package_path,
                root_package_name='abjad',
                )
            for class_ in generator:
                if (
                    issubclass(class_, base_class) and
                    class_ is not base_class and
                    not inspect.isabstract(class_)
                    ):
                    script_classes.append(class_)
        return list(sorted(script_classes, key=lambda x: x.__name__))

    ### PUBLIC PROPERTIES ###

    @property
    def argument_parser(self):
        r'''The script's instance of argparse.ArgumentParser.
        '''
        return self._argument_parser

    @property
    def formatted_help(self):
        r'''Formatted help of developer script.
        '''
        return self._argument_parser.format_help()

    @property
    def formatted_usage(self):
        r'''Formatted usage of developer script.
        '''
        return self._argument_parser.format_usage()

    @property
    def formatted_version(self):
        r'''Formatted version of developer script.
        '''
        return self._argument_parser.format_version()

    @property
    def program_name(self):
        r'''The name of the script, callable from the command line.
        '''
        name = type(self).__name__[:type(self).__name__.rfind('Script')]
        return stringtools.to_space_delimited_lowercase(name).replace(' ', '-')



# -*- coding: utf-8 -*-
from abjad import *


def test_indicatortools_Dynamic_is_dynamic_name_01():

    assert Dynamic.is_dynamic_name('ff')
    assert Dynamic.is_dynamic_name('f')
    assert Dynamic.is_dynamic_name('p')
    assert Dynamic.is_dynamic_name('pp')


def test_indicatortools_Dynamic_is_dynamic_name_02():

    assert not Dynamic.is_dynamic_name('x')



# -*- coding: utf-8 -*-
import pytest
from abjad import *


def test_spannertools_Spanner__get_my_nth_leaf_01():

    staff = Staff()
    staff.append(Measure((2, 8), "c'8 d'8"))
    staff.append(Measure((2, 8), "e'8 f'8"))
    leaves = select(staff).by_leaf()
    beam = Beam()
    attach(beam, leaves)

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            {
                \time 2/8
                c'8 [
                d'8
            }
            {
                e'8
                f'8 ]
            }
        }
        '''
        )

    selector = select().by_leaf(flatten=True)
    leaves = selector(staff)

    assert beam._get_my_nth_leaf(0) is leaves[0]
    assert beam._get_my_nth_leaf(1) is leaves[1]
    assert beam._get_my_nth_leaf(2) is leaves[2]
    assert beam._get_my_nth_leaf(3) is leaves[3]

    assert beam._get_my_nth_leaf(-1) is leaves[-1]
    assert beam._get_my_nth_leaf(-2) is leaves[-2]
    assert beam._get_my_nth_leaf(-3) is leaves[-3]
    assert beam._get_my_nth_leaf(-4) is leaves[-4]

    assert pytest.raises(IndexError, 'beam._get_my_nth_leaf(99)')
    assert pytest.raises(IndexError, 'beam._get_my_nth_leaf(-99)')



# -*- coding: utf-8 -*-


def all_are_unequal(argument):
    '''Is true when `argument` is an iterable collection of unequal items.
    Otherwise false.

    ..  container:: example

        ::

            >>> mathtools.all_are_unequal([1, 2, 3, 4, 9])
            True

        ::

            >>> mathtools.all_are_unequal(17)
            False

    ..  container:: example

        Is true when `argument` is empty:

        ::

            >>> mathtools.all_are_unequal([])
            True

    Returns true or false.
    '''
    try:
        return argument == type(argument)(set(argument))
    except TypeError:
        return False
